from typing import Any

from spread_core import mqtt
from spread_core.errors.project_errors import ClassifierError, CommandError
from spread_core.mqtt.spread.topic import Set, Renew
from spread_core.mqtt.spread.variable import Variable
from spread_core.tools import settings
from spread_core.tools.signature import CommandSignature
from spread_core.tools.utils import get_subclass


class EntityMeta(type):
    @property
    def entity_type(cls):
        return cls.__name__


class Entity(metaclass=EntityMeta):
    funit_class = None
    gettable_units = None
    settable_units = None

    def __init__(self, id: int, name='', **kwargs):
        self._id = id
        self.values = dict()
        self._invalid_values = {}
        self._preliminaries = []
        self._inconsistent = []
        self._invalid = []
        self.name = name or f'id:{id}'
        self.entity_type = type(self).entity_type

    @staticmethod
    def parse_args(**kwargs) -> dict:
        """ Преобразует атрибуты проекта в параметры класса """
        return {}

    @classmethod
    def get_address_broadcast(cls, p_id, funit_type='+'):
        raise NotImplementedError()

    @property
    def id(self):
        return self._id

    @property
    def project_id(self) -> int:
        raise NotImplementedError()

    @property
    def rank(self):
        return self.id

    def __str__(self): return '{}({})'.format(self.__class__.__name__, self.id)
    def __repr__(self): return self.__str__()

    def is_dumped(self, key) -> bool: return False
    def get_address(self, funit_type): raise NotImplementedError()

    def is_repaired_funit(self, funit_type) -> bool:
        """ Whether to restore the stored value when restoring the bus or starting the service """
        return True

    def publish(self, topic, data, retain=False, **kwargs):
        raise NotImplemented()

    def get_info(self):
        pass

    def on_command(self, topic: (Set, Renew), variable: Variable):
        funit_type = topic.entity_address.funit_type
        command = self.funit_class.of(funit_type)
        if command and issubclass(command, Funit):
            try:
                if isinstance(topic, Set):
                    command(self, variable).set()
                elif isinstance(topic, Renew):
                    command(self, variable).get()
                else:
                    raise ClassifierError.of_action(topic.__class__.__name__, self, funit_type)
            except BaseException as ex:
                raise CommandError.of_exec(str(ex))
        else:
            raise ClassifierError.of_funit(funit_type, self)

    def on_update(self, funit_type, value, sig=None, **kwargs):
        correction = False
        preliminary = kwargs.pop(Variable.PRELIM, False)
        variable = Variable(value, preliminary=preliminary)

        if funit_type in self._preliminaries:
            if not preliminary:
                correction = True
                self._preliminaries.remove(funit_type)

        if value != self.values.get(funit_type, None) or correction:
            if preliminary and value is not None:
                self._preliminaries.append(funit_type)

            self.make_state(funit_type, variable, **kwargs)

        if sig:
            self.make_reply(funit_type, variable, sig)

    def make_state(self, funit_type, variable, **kwargs):
        if variable.invalid:
            has_value = self.values.pop(funit_type, None)
            if has_value is not None and self.is_repaired_funit(funit_type):
                self._invalid_values[funit_type] = has_value
        else:
            self.set_value(funit_type, variable.value)
            self._invalid_values.pop(funit_type, None)

        self.publish(
            topic=mqtt.spread.topic.State(self.get_address(funit_type)),
            data=variable,
            retain=isinstance(funit_type, str) or funit_type.retained,
            **kwargs
        )

    def make_reply(self, funit_type, variable, sig):
        if sig and sig.session_id and sig.funit_type == funit_type:
            if sig.key:
                variable.key = sig.key
            self.publish(mqtt.spread.topic.Reply(self.project_id, sig.session_id), variable, retain=False)

    def get_value(self, funit_type, def_value=None):
        if funit_type in self.values:
            return self.values[funit_type]
        else:
            _exist, _ow = settings.dump.get_dump(self.__class__.__name__, self.id, funit_type)
            if _exist:
                return _ow
            else:
                return def_value

    def set_value(self, key, value):
        self.values[key] = value
        if self.is_dumped(key):
            if isinstance(key, FunitMetaClass) or (isinstance(key, type) and issubclass(key, FunitMetaClass)):
                key = key.funit_type
            settings.dump.set_dump(self, key, value)


class FunitMetaClass(type):
    @property
    def funit_type(cls): return cls.__name__

    def __eq__(cls, other):
        if isinstance(other, FunitMetaClass):
            return other.funit_type == cls.funit_type

        return other == cls.funit_type

    def __str__(cls): return cls.funit_type
    def __hash__(cls): return hash(cls.funit_type)


class Funit(metaclass=FunitMetaClass):
    retained = True
    delay = 0
    entity_type = Entity
    variants = None
    mapped_value = None
    index_of = []
    mask = None
    value_type = None
    value_range = None
    hex_count = 0
    multiplier = 0
    round_to = None

    class NotImplemented(Exception):
        pass

    def __init__(self, entity: entity_type, variable: [Variable, Any]):
        self._entity = entity
        self._variable = variable if isinstance(variable, Variable) else Variable(variable)
        self._set_value = None

    @classmethod
    def of(cls, funit_type): return get_subclass(cls, lambda cmd: issubclass(cmd, cls) and cmd.funit_type == funit_type)

    @property
    def funit_type(self): return self.__class__.funit_type

    @property
    def entity(self) -> entity_type: return self._entity

    @property
    def signature(self) -> [CommandSignature, None]:
        if self.signed:
            return CommandSignature(self.funit_type, session_id=self._variable.session_id, key=self._variable.key)

        return None

    @classmethod
    def of_raw(cls, raw_value):
        if cls.multiplier:
            raw_value /= int(cls.multiplier)

        if cls.round_to:
            raw_value = round(raw_value, cls.round_to)

        if cls.index_of:
            return cls.index_of[raw_value]
        elif cls.mask:
            return [cls.mask[i] for i in range(0, 8) if (raw_value & (2 ** i))]
        elif cls.hex_count:
            return hex(raw_value)[2:].upper().rjust(cls.hex_count*2, '0')

        return raw_value

    @property
    def set_value(self):
        if self._set_value is None:
            _value = self._variable.value
            if self.multiplier:
                _value *= self.multiplier

            if self.hex_count:
                _value = int(_value, 16)

            if self.round_to:
                _value = round(_value, self.round_to)

            if self.index_of:
                if _value not in self.index_of:
                    raise ValueError('Unexpected value: %s' % _value)
                _value = self.index_of.index(_value)
            elif self.variants:
                if _value not in self.variants:
                    raise ValueError('Unexpected value: %s' % _value)
                if self.mapped_value:
                    _value = self.mapped_value[self.variants.index(_value)]
            elif self.mask:
                try:
                    _value = sum([1 << self.mask.index(i) for i in set(_value)])
                except:
                    raise ValueError('Unexpected value: %s' % _value)
            elif self.value_range and _value is not None and int(_value) not in self.value_range:
                raise ValueError('Unexpected value: %s' % _value)

            if self.value_type:
                try:
                    _value = self.value_type(_value)
                except:
                    raise ValueError('Unexpected value type: %s' % _value)

            self._set_value = _value

        return self._set_value

    @property
    def params(self) -> dict: return self._variable.params

    @property
    def signed(self): return self._variable.session_id

    def set(self): raise self.NotImplemented()
    def get(self): raise self.NotImplemented()
    def state(self): raise self.NotImplemented()
