import json
import logging
import queue
import threading

import spread_core.errors.project_errors as errors
import spread_core.mqtt as mqtt
import spread_core.tools.settings_const as const
from spread_core.bam.providers import Provider
from spread_core.bam.survey import Survey
from spread_core.tools.base_interface import AbstractInterfaceClass
from spread_core.tools.corutunes import coroutine
from spread_core.tools.manager_interface import ManagerOfBroker
from spread_core.tools.settings import config, POLL_RATE
from spread_core.warehouse.warehouse_interface import WarehouseInterface
from .entities import Entity, Funit

DEF_POLL_RATE = 0


class Manager(Entity, ManagerOfBroker, WarehouseInterface):
    STATUS_STOPPED = 'Stopped'
    STATUS_RUNNING = 'Running'
    STATUS_INFO = 'Info'
    STATUS_SURVEY = 'Survey'

    def __init__(self, p_id,  m_id):
        Entity.__init__(self, m_id)
        WarehouseInterface.__init__(self)
        self._def_poll_rate = DEF_POLL_RATE
        self._stopped = False
        self._send_interface = None
        self._project_id = p_id
        self.poll_rate = 0
        self.state_accumulator_event = threading.Event()
        self.bus = None
        self.stack = queue.Queue()
        self.has_inited = False
        self.surver = Survey(self)

        self.update_config()

    @property
    def send_interface(self) -> AbstractInterfaceClass:
        return self._send_interface

    @send_interface.setter
    def send_interface(self, interface):
        if self._send_interface:
            self._send_interface.set_stopped()
            del self._send_interface

        self._send_interface = interface

        if interface:
            interface.set_event_handler(self.event_handler())
            interface.update_config()
            interface.set_external_cmd_handler(self.external_command_handler())
            interface.set_status_handler(self.bus_status_handler())
            interface.set_bus_params(self.bus)
            interface.start()

    @property
    def project_id(self) -> int:
        return self._project_id

    def publish(self, topic, data, retain=False, **kwargs):
        ManagerOfBroker.publish(self, topic, data, retain=retain, **kwargs)

    def parse(self, data):
        attributes = data[const.ATTRIBUTES]

        self.bus = attributes.get('bus', None)
        self.name = data.get('name', f'id:{self.id}')
        self._def_poll_rate = attributes.get(POLL_RATE, DEF_POLL_RATE)
        self.update_poll_rate()

    def set_broker(self, mqttc):
        ManagerOfBroker.__init__(self, mqttc)

    def set_interface(self):
        raise NotImplementedError()

    def update_config(self):
        self.update_poll_rate()
        if self._send_interface:
            self._send_interface.update_config()

    def update_poll_rate(self):
        pl = config.get(POLL_RATE, None)

        if pl is None:
            pl = self._def_poll_rate

        try:
            pl = int(pl)
        except ValueError:
            logging.warning('Invalid "pollRate" value: %s. Defaulted <%s> will used' % (pl, DEF_POLL_RATE))
            pl = DEF_POLL_RATE

        if pl:
            pl = max(0.1, round(pl / 1000, 1))

        self.poll_rate = pl

    def start(self):
        self.subscribe(mqtt.project.ProjectFile.managers(self.project_id))

    def on_state(self, topic: mqtt.spread.topic.SpreadTopic = None, variable: mqtt.spread.variable.Variable = None):
        if topic and variable:
            if isinstance(topic.entity_address, mqtt.spread.address.ProviderAddress):
                _provider = self.get_provider(topic.entity_address.id)
                if _provider:
                    _provider.on_state(topic.entity_address.funit_type, variable)

        if not self.state_accumulator_event.is_set():
            self.state_accumulator_event.set()

    def on_project(self, topic, data):
        self.unsubscribe(topic)

        if isinstance(data, (bytes, bytearray)):
            data = data.decode()

        payload = json.loads(data)
        if mqtt.project.ProjectFile.MANAGERS in payload:
            self.on_project_managers(payload[mqtt.project.ProjectFile.MANAGERS])

        elif mqtt.project.ProjectFile.PROVIDERS in payload:
            self.on_project_providers(payload[mqtt.project.ProjectFile.PROVIDERS])

    def on_project_managers(self, data: list):
        for item in data:
            if item.get('id', None) == self.id:
                try:
                    self.parse(item)
                except Exception as ex:
                    logging.exception(ex)
                    raise errors.InitError("Invalid project data %s(%s)" % (self.__class__.__name__, self.id))

                self.subscribe(mqtt.project.ProjectFile.providers(self.project_id), True)
                break
        else:
            raise errors.InitError("%s(%s) absent in project" % (self.__class__.__name__, self.id))

    def on_project_providers(self, data: list):
        for item in data:
            _id = item.pop('id', None)
            _type = item.pop('type', None)

            if _id is None or _type is None:
                logging.warning(f'Empty provider data of {_type}({_id})')
            else:
                self.on_project_provider(_id, _type, item)

        logging.info('Project({}) loaded'.format(self.project_id))
        self.subscribe(mqtt.spread.topic.State(self.get_address('+/+/#')))
        threading.Thread(target=self.on_ready, name="Broker").start()

    def on_project_provider(self, entity_id: int, entity_type: str, data: dict):
        from spread_core.bam.generator import generate_provider

        if data.pop('managerID', None) == self.id:
            try:
                provider = generate_provider(entity_id, entity_type, data)
                self.add_provider(provider)
            except BaseException as ex:
                logging.warning(ex)

    def on_ready(self):
        while self.state_accumulator_event.wait(1):
            self.state_accumulator_event.clear()

        if self._stopped:
            return

        self.unsubscribe(mqtt.spread.topic.State(self.get_address('+/+/#')))
        self.subscribe(mqtt.spread.topic.Set(self.get_address(None)))
        self.subscribe(mqtt.spread.topic.Renew(self.get_address(None)))

        try:
            self.set_interface()
            self.info()
        except Exception as ex:
            logging.error(ex)

    def on_exit(self):
        super(Manager, self).on_exit()
        self._stopped = True

        self.state_accumulator_event.set()

        if self.poll_rate != 0 and self.stack.qsize() == 0:
            self.stack.put(None)

        if self._send_interface:
            self._send_interface.set_stopped()

        if self.mqttc.connected:
            for _provider in self.get_all_devices():
                _provider.on_exit()

    def survey(self):
        for provider in self.get_all_devices():
            if provider.surveyable:
                yield from provider.on_survey()

    def status_survey(self):
        for provider in self.get_all_devices():
            yield from provider.on_status_survey()

    @coroutine
    def external_command_handler(self):
        while True:
            try:
                ext = yield
                try:
                    self.on_external_command(*ext)
                except BaseException as ex:
                    logging.debug(f'EXTERNAL FRAME ERROR[{ext}]')
                    logging.exception(ex)
            except StopIteration:
                break

    def on_external_command(self, *args):
        raise NotImplementedError()

    @coroutine
    def event_handler(self):
        while True:
            try:
                event = yield
                try:
                    self.on_event(event)
                except BaseException as ex:
                    logging.debug(f'EVENT FRAME ERROR[{event}]')
                    logging.exception(ex)
            except StopIteration:
                break

    def on_event(self, event):
        raise NotImplementedError()

    @coroutine
    def bus_status_handler(self):
        while True:
            ready = yield
            if self.has_inited:
                self.on_bus_status(ready)

    def on_bus_status(self, ready):
        logging.info('Bus status is %s' % ready)
        for p in self.get_all_devices():
            if isinstance(p, Provider):
                p.set_valid(p.VALID if ready else p.INVALID)

    def pre_info(self):
        while not self.send_interface.ready:
            if self._stopped:
                break

            threading.Event().wait(1)

    def info(self):
        self.pre_info()

        if self._stopped:
            return

        for provider in self.sorted_info():
            if self._stopped:
                return
            if provider.infoble:
                try:
                    provider.check_valid()
                    if provider.is_valid:
                        provider.restore_states()
                        provider.get_info()
                        provider.override_info()
                except errors.InvalidDevice as ex:
                    logging.warning(ex)
                    provider.set_valid(provider.INVALID)

        self.post_info()

    def post_info(self):
        self.has_inited = True
        self.do_next()

    def do_next(self):
        while not self._stopped:
            try:
                command = self.stack.get(timeout=self.poll_rate or None)
                if command:
                    command.__call__()
            except queue.Empty:
                self.surver.next()
            except BaseException as ex:
                logging.exception(ex)
            else:
                self.stack.task_done()

    def put_to_stack(self, funit_func):
        self.stack.put(funit_func)

    def on_command(self, topic: mqtt.spread.topic.Set, variable: mqtt.spread.variable.Variable):
        self.log_command(topic, variable)
        if self._stopped:
            return

        try:
            if topic.entity_address.type == self.__class__.__name__ and topic.entity_address.id == self.id:
                super(Manager, self).on_command(topic, variable)
            elif not self._send_interface.ready:
                raise errors.JocketError('Bus is not ready')
            else:
                _provider = self.get_provider(topic.entity_address.id)
                if _provider:
                    if _provider.is_valid:
                        command = _provider.funit_class.of(topic.entity_address.funit_type)
                        if not command:
                            command = Funit.of(topic.entity_address.funit_type)
                        if command:
                            command = command(_provider, variable)

                            if isinstance(topic, mqtt.spread.topic.Set):
                                command = command.set
                            elif isinstance(topic, mqtt.spread.topic.Renew):
                                command = command.get
                            else:
                                raise Funit.NotImplemented()

                            self.stack.put(command)
                        else:
                            raise errors.CommandError('Unknown funit')
                    else:
                        raise errors.InvalidDevice(_provider)
                else:
                    raise errors.JocketError('Объект с адресом {} не найден!'.format(topic.entity_address.id))
        except (errors.JocketError, errors.ClassifierError, errors.InvalidDevice, errors.CommandError) as ex:
            logging.warning(ex)
            if variable.session_id:
                to_topic = mqtt.spread.topic.Reply(self.project_id, variable.session_id)
                self.publish(to_topic, mqtt.spread.variable.Variable(None, error=str(ex)))

    def log_command(self, topic: mqtt.spread.topic.SpreadTopic, variable: mqtt.spread.variable.Variable):
        logging.debug(
            f'COMMAND({variable.session_id if variable.session_id else ""}): ' +
            f'{topic.__class__.__name__} {topic.entity_address.type}({topic.entity_address.id})'
            f'.{topic.entity_address.funit_type}' +
            (f'={variable.value}' if isinstance(topic, mqtt.spread.topic.Set) else '') +
            (f' params: {variable.params}' if variable.params else '') +
            f' queue: {self.stack.qsize()}'
        )

    def get_address(self, funit_type) -> mqtt.spread.address.ManagerAddress:
        return mqtt.spread.address.ManagerAddress(
            self._project_id,
            self.__class__.__name__,
            self.id,
            funit_type or "#"
        )
