# import json
import re
from spread_core.bam.validation.base import corr_par, corr_req_par, group_address
import spread_core.bam.validation.constraints as constr


class ProviderValidator:
    def __init__(self, data):
        if not isinstance(data, dict):
            raise ValueError('Wrong data structure (dict needed)')
        if not corr_req_par('type', str, data):
            raise ValueError('Wrong type')

        in_constr = False
        for man in constr.MAN_PROV_CONSTR:
            if data['type'] in constr.MAN_PROV_CONSTR[man]:
                in_constr = True
                break
        if not in_constr:
            raise ValueError('Wrong type')

        if not corr_req_par('id', int, data) or data['id'] < 1:
            raise ValueError('Wrong id')
        # key must be GUID
        if not corr_req_par('key', str, data) \
                or not re.match(r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', data['key']):
            raise ValueError('Wrong key')
        if not corr_req_par('managerID', int, data) or data['managerID'] < 1:
            raise ValueError('Wrong managerID')
        if not corr_par('name', str, data):
            raise ValueError('Wrong name')
        self.Attributes(data)

    def Attributes(self, data):
        error = f"Wrong {data['type']} attribute"
        if data['type'] in ('TeletaskDimmer', 'TeletaskRelay', 'TeletaskTemperatureSensor', 'TeletaskThermoregulator', 'TeletaskMotor'):
            if 'attributes' not in data:
                raise ValueError(f"{error}s")
            if not corr_req_par('controller', int, data['attributes']) \
                    or data['attributes']['controller'] < 1 or data['attributes']['controller'] > 10:
                raise ValueError(f"{error} controller")
            if not corr_req_par('channel', int, data['attributes']) \
                    or data['attributes']['channel'] < 0 or data['attributes']['channel'] > 65535:
                raise ValueError(f"{error} channel")

        elif data['type'] in ('DaliDimmer', 'DaliRelay', 'DaliTunableWhite', 'DaliLightSensor', 'DaliPresenceSensor', 'DaliPushButton', 'DaliRgbwaf'):
            if 'attributes' not in data:
                raise ValueError(f"{error}s")
            if not corr_req_par('binding', str, data['attributes']) \
                    or data['attributes']['binding'] not in ('Device', 'Group', 'Broadcast'):
                raise ValueError(f"{error} binding")

            if data['attributes']['binding'] == 'Device':
                if not corr_req_par('device', int, data['attributes']) \
                        or data['attributes']['device'] < 0 or data['attributes']['device'] > 63:
                    raise ValueError(f"{error} device")

            if data['attributes']['binding'] == 'Group':
                if not corr_req_par('group', int, data['attributes']) \
                        or data['attributes']['group'] < 0 or data['attributes']['group'] > 15:
                    raise ValueError(f"{error} group")

            if data['type'] == 'DaliPushButton':
                if not corr_req_par('instanceIndex', int, data['attributes']) \
                        or data['attributes']['instanceIndex'] < 0 or data['attributes']['instanceIndex'] > 31:
                    raise ValueError(f"{error} instanceIndex")

        elif data['type'] == 'MercuryElectricMeter':
            if 'attributes' not in data:
                raise ValueError(f"{error}s")
            if not corr_req_par('address', int, data['attributes']) \
                    or data['attributes']['address'] < 0 or data['attributes']['address'] > 255:
                raise ValueError(f"{error} address")
            if not corr_req_par('password1', str, data['attributes']):
                raise ValueError(f"{error} password1")
            if not corr_req_par('password2', str, data['attributes']):
                raise ValueError(f"{error} password2")

        elif data['type'] == 'LomDimmer':
            if 'attributes' not in data:
                raise ValueError(f"{error}s")
            if not corr_req_par('devEui', str, data['attributes']) \
                    or not re.match(r'[0-9A-F]{16}$', data['attributes']['devEui']):
                raise ValueError(f"{error} devEui")

        elif data['type'] == 'LomGroup':
            if 'attributes' not in data:
                raise ValueError(f"{error}s")
            if not corr_req_par('address', str, data['attributes']) \
                    or not re.match(r'[0-9A-F]+$', data['attributes']['address']):
                raise ValueError(f"{error} address")

        elif data['type'] in ('EwsRoom', 'EwsEquipment'):
            if 'attributes' not in data:
                raise ValueError(f"{error}s")
            if not corr_req_par('alias', str, data['attributes']):
                raise ValueError(f"{error} alias")

        elif data['type'] in ('KnxDimmer', 'KnxRelay'):
            if 'attributes' not in data:
                raise ValueError(f"{error}s")
            if 'state' not in data['attributes'] or not group_address(data['attributes']['state']):
                raise ValueError(f"{error} state")
            if 'control' not in data['attributes'] or not group_address(data['attributes']['control']):
                raise ValueError(f"{error} control")

        elif 'attributes' in data:
            raise ValueError(f"{error}s")
