import os
import sys

from ..base import Topic, TopicError


LOCAL = 'Local'
HASH = 'Hash'
CTP = 'Ctp'
FILE = 'File'
SET = 'Set'
RESET = 'Reset'

PRESETS        = "Preset"
EQUIPMENT      = "Equipment"
LOCATION       = "Location"
HARDWARE       = "Hardware"
HEADER         = "Header"
USER           = "User"
CYCLE          = "Cycle"
TRIGGERS       = "Triggers"


def os_dependence(fn):
    def wrap(*args, **kwargs):
        result = fn(*args, **kwargs)
        return os.path.join(*result.split('/'))

    if sys.platform == 'win32':
        return wrap

    return fn


class ProjectTopic(Topic):
    protocol = 'Project'

    def __init__(self, project_id: int):
        self.project_id = int(project_id)

    def __str__(self):
        return f'{self.protocol}/{self.project_id}'

    @classmethod
    def broadcast(cls, project_id, *parts):
        return f'{cls.protocol}/{project_id}/{"/".join(parts)}'

    @classmethod
    def decompose(cls, project_id, sub_protocol, *parts):
        if sub_protocol == LOCAL:
            return Local.decompose(project_id, *parts)

        elif sub_protocol == HASH:
            return Hash(project_id)

        elif sub_protocol == CTP:
            return ProjectFile(project_id, sub_protocol)

        elif sub_protocol == FILE:
            return ProjectFile(project_id, *parts)

        elif sub_protocol == SET:
            return Set(project_id)

        elif sub_protocol == RESET:
            return Reset(project_id)

        else:
            raise TopicError('Unknown family: "%s"' % sub_protocol)


class Hash(ProjectTopic):

    def __str__(self):
        return f'{self.protocol}/{self.project_id}/{HASH}'


class Set(ProjectTopic):

    def __str__(self):
        return f'{self.protocol}/{self.project_id}/{SET}'

    @classmethod
    def broadcast(cls, project_id, *parts):
        return f'{cls.protocol}/{project_id}/{SET}/{"/".join(parts)}'


class Reset(ProjectTopic):

    def __str__(self):
        return f'{self.protocol}/{self.project_id}/{RESET}'

    @classmethod
    def broadcast(cls, project_id, *parts):
        return f'{cls.protocol}/{project_id}/{RESET}/{"/".join(parts)}'


class ProjectFile(ProjectTopic):
    ENGINERIES = 'engineries'
    PROVIDERS = 'providers'
    MANAGERS = 'managers'
    SUBGINERIES = 'subgineries'
    HEADER = 'header'
    SERVERS = 'servers'
    LOCATIONS = 'locations'
    MODELS = 'models'
    CTP = 'Ctp'
    USERS = 'users'

    FILE = 'File'

    def __init__(self, project_id, file_name):
        super(ProjectFile, self).__init__(project_id)
        self.file_name = file_name

    def __str__(self):
        s = f'{self.protocol}/{self.project_id}'

        if self.file_name != CTP:
            s += f'/{FILE}'

        s += f'/{self.file_name}'

        return s

    @classmethod
    def engineries(cls, p_id): return cls(p_id, f'{ProjectFile.ENGINERIES}.json')

    @classmethod
    def subgineries(cls, p_id): return cls(p_id, f'{ProjectFile.SUBGINERIES}.json')

    @classmethod
    def providers(cls, p_id): return cls(p_id, f'{ProjectFile.PROVIDERS}.json')

    @classmethod
    def managers(cls, p_id): return cls(p_id, f'{ProjectFile.MANAGERS}.json')

    @classmethod
    def servers(cls, p_id): return cls(p_id, f'{ProjectFile.SERVERS}.json')

    @classmethod
    def header(cls, p_id): return cls(p_id, f'{ProjectFile.HEADER}.json')

    @classmethod
    def locations(cls, p_id): return cls(p_id, f'{ProjectFile.LOCATIONS}.json')

    @classmethod
    def ctp(cls, p_id): return cls(p_id, ProjectFile.CTP)


class Local(ProjectTopic):
    family = None

    def __init__(self, project_id, entity_type, entity_id):
        super(Local, self).__init__(project_id)

        self.suffix = f'{self.protocol}/{project_id}'

        self._entity_type = entity_type
        self._entity_id = int(entity_id)

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return f'{self.suffix}/{self.rootfix()}/{self.postfix}'

    def __hash__(self):
        return hash(str(self))

    @property
    def entity_type(self) -> str:
        """ Type of target entity """
        return self._entity_type

    @property
    def entity_id(self) -> int:
        """ ID of target entity """
        return self._entity_id

    @classmethod
    def suffix(cls, project_id) -> str:
        """ Returns left part of topic like 'Project/<project_id>/...'"""
        return f'{cls.protocol}/{project_id}'

    @classmethod
    def rootfix(cls) -> str:
        """ Returns middle part of topic like ".../Local/<family>/..." """
        return f'{LOCAL}/{cls.family}'

    @property
    def postfix(self) -> str:
        """ Returns right part of topic like '.../<entity_type>/<entity_id>'"""
        if self._entity_id == -999:
            return f'{self._entity_type}'

        return f'{self._entity_type}/{self._entity_id}'

    @os_dependence
    def path(self, root: str) -> str:
        """ Returns destination of corresponding file as "[<root>/]path/to/file.json" """
        root = root.split(str(self.project_id))[0]
        return os.extsep.join([os.path.join(self.destination(root, self.project_id), self.postfix), 'json'])

    @classmethod
    @os_dependence
    def destination(cls, root: str, project_id: int) -> str:
        """ Returns destination of corresponding dir as "[<root>/]<project_id>/Local/<family>" """
        root = root.split(str(project_id))[0]
        return os.path.join(root, str(project_id), cls.rootfix())

    @classmethod
    def of_file(cls, project_id: int, dest, file) -> str:
        """ Returns the topics string that corresponds to file """
        dest = dest[1:] if dest.startswith(os.path.sep) else dest
        file = os.path.splitext(file)[0]
        return f'{cls.suffix(project_id)}/{LOCAL}/{dest}/{file}'

    @classmethod
    def broadcast(cls, project_id, entity_type='+', entity_id='+'):
        """ Returns the topics string that corresponds to entity with arbitrary parts """
        if entity_type == '#':
            if cls.family is None:
                return f'{cls.suffix(project_id)}/Local/{entity_type}'
            else:
                return f'{cls.suffix(project_id)}/{cls.rootfix()}/{entity_type}'

        return f'{cls.suffix(project_id)}/{cls.rootfix()}/{entity_type}/{entity_id}'

    @classmethod
    def decompose(cls, project_id, family, *parts):
        if family == PRESETS:
            return Preset(project_id, *parts)

        elif family == EQUIPMENT:
            return Equipment(project_id, *parts)

        elif family == LOCATION:
            if len(parts) == 1:
                return Location(project_id, *parts)

            return Subginery(project_id, *parts)

        elif family == HARDWARE:
            if len(parts) == 2:
                return Hardware(project_id, *parts)

            return Provider(project_id, *parts)

        elif family == HEADER:
            return Header(project_id, *parts)

        elif family == USER:
            return User(project_id, *parts)

        elif family == CYCLE:
            if len(parts) == 3:
                if parts[-2] == TRIGGERS:
                    return CycleTrigger(project_id, *parts[::2])
            else:
                return Cycle(project_id, *parts)

        raise TopicError('Unknown family: "%s"' % family)


class _LocalOnlyID(Local):
    def __init__(self, project_id: int,  entity_id: int):
        super(_LocalOnlyID, self).__init__(project_id,  None, entity_id)

    @property
    def postfix(self):
        return f'{self._entity_id}'

    @classmethod
    def broadcast(cls, project_id='+', entity_id='+', *_):
        return f'{cls.suffix(project_id)}/{cls.rootfix()}/{entity_id}'


class Preset(Local):
    family = PRESETS

    @classmethod
    def destination(cls, root: str, project_id: int, entity_type: str = None) -> str:
        destination = super(Preset, cls).destination(root, project_id)
        if entity_type:
            return os.path.join(destination, entity_type)

        return destination


class Equipment(Local):
    family = EQUIPMENT


class Location(_LocalOnlyID):
    family = LOCATION


class Subginery(Local):
    family = LOCATION

    def __init__(self, project_id: int,  location_id: int, entity_type, entity_id: int):
        super(Subginery, self).__init__(project_id, entity_type, entity_id)
        self._location_id = int(location_id)

    @property
    def location_id(self) -> int:
        return self._location_id

    @property
    def postfix(self):
        return f'{self._location_id}/{self._entity_type}/{self._entity_id}'

    @classmethod
    def broadcast(cls, project_id='+', location_id='+', entity_type='+', entity_id='+'):
        return f'{cls.suffix(project_id)}/{cls.rootfix()}/{location_id}/{entity_type}/{entity_id}'


class Hardware(Local):
    family = HARDWARE


class Provider(Local):
    family = HARDWARE

    def __init__(self, project_id: int,  owner_type, owner_id: int, entity_type, entity_id: int):
        super(Provider, self).__init__(project_id, entity_type, entity_id)

        self._owner_type = owner_type
        self._owner_id = int(owner_id)

    @property
    def manager_type(self) -> str:
        return self._owner_type

    @property
    def manager_id(self) -> int:
        return self._owner_id

    @property
    def postfix(self):
        return f'{self._owner_type}/{self._owner_id}/{self._entity_type}/{self._entity_id}'

    @classmethod
    def broadcast(cls, project_id='+', owner_type='+', owner_id='+', entity_type='+', entity_id='+'):
        return f'{cls.suffix(project_id)}/{cls.rootfix()}/{owner_type}/{owner_id}/{entity_type}/{entity_id}'


class Header(Local):
    family = HEADER

    def __init__(self, project_id: int,  entity_type: str):
        super(Header, self).__init__(project_id, entity_type, -999)

    @property
    def postfix(self):
        return f'{self._entity_type}'

    @classmethod
    def broadcast(cls, project_id='+', entity_type='+', *_):
        return f'{cls.suffix(project_id)}/{cls.rootfix()}/{entity_type}'


class User(_LocalOnlyID):
    family = USER


class Cycle(_LocalOnlyID):
    family = CYCLE

    @os_dependence
    def path(self, root: str) -> str:
        """ Returns destination of corresponding file as "[<root>/]path/to/file" """
        root = root.split(str(self.project_id))[0]
        return os.path.join(self.destination(root, self.project_id), self.postfix)


class CycleTrigger(Local):
    family = CYCLE

    def __init__(self, project_id: int,  cycle_id: int, entity_id: int):
        super(CycleTrigger, self).__init__(project_id, TRIGGERS, entity_id)

        self._cycle_id = int(cycle_id)

    @property
    def cycle_id(self):
        return self._cycle_id

    @property
    def postfix(self):
        return f'{self._cycle_id}/{TRIGGERS}/{self._entity_id}'

    @classmethod
    def broadcast(cls, project_id='+', cycle_id='+', entity_id='+'):
        return f'{cls.suffix(project_id)}/{cls.rootfix()}/{cycle_id}/{TRIGGERS}/{entity_id}'
