import asyncio
import sys
import json
import logging
import os
import time
import zipfile
import zlib

from asyncio_mqtt import Client, MqttError

from spread_core import mqtt
from spread_core.tools import settings, utils
import spread_lom as lom
from spread_core.bam.engineries.lighting_area import ButtonPreset
from spread_core.bam.validation import enginery, manager, provider, subginery, location, header_firmware, \
                                       header_project, cycle, button_preset, trigger, user
import spread_core.bam.validation.constraints as constr
from spread_core.projector.base import FileSystemStorage, file_sys_storage_modify, VerifyError

start_time = None

# ENG_CONSTR =      {'enginery_type':  {'recipe_type': {'ingredient_type': ['provider_type']}}}
# MAN_PROV_CONSTR = {'manager_type': ['provider_type']}
# SUBG_CONSTR =     {'subginery_type': ['enginery_type']}


class Launcher:
    """ Abstract launcher """

    def stop(self): raise NotImplemented()

    def update(self): raise NotImplemented()


class Projector(Launcher):
    def __init__(self):
        self.client = None
        self.storage = FileSystemStorage(settings.config[settings.LOCAL_PATH], settings.config[settings.PROJECT_ID])
        self.verify_mode = settings.config.get(settings.VERIFY_MODE, 0)

        if not isinstance(self.verify_mode, int) or self.verify_mode not in range(0, 5):
            logging.error("Service can't start. Incorrect 'Verify mode' param in service config.")
            sys.exit(0)

        self.SERVERS = []         # [server_id]
        self.LOCATIONS = []       # [location_id]
        self.CYCLES = []          # [cycle_id]
        self.PROVIDERS = {}       # {'provider_id': 'provider_type'}
        self.MANAGERS = {}        # {'manager_id': 'manager_type'}
        self.ENGINERIES = {}      # {'enginery_id': 'enginery_type'}
        self.BROKER_REVISION_STAGE = True
        self.ctp = {}
        self.project = {}

    @property
    def project_id(self): return settings.config[settings.PROJECT_ID]

    @property
    def ctp_path(self): return settings.config[settings.CTP_PATH]

    def stop(self):
        logging.debug('Stop service')
        raise KeyboardInterrupt()

    def update(self):
        self.BROKER_REVISION_STAGE = True
        logging.debug('Restart service')
        raise MqttError()

    async def subscribe(self, topic, logged=True):
        await self.client.subscribe(str(topic))
        if logged:
            logging.debug('Subscribed to %s' % topic)

    async def unsubscribe(self, topic, logged=True):
        await self.client.unsubscribe(str(topic))
        if logged:
            logging.debug('Unsubscribed from %s' % topic)

    async def publish(self, topic: (str, mqtt.Topic), data, retain=True, logged=True):
        if not isinstance(data, (int, str, bytes, bytearray)):
            data = json.dumps(data)

        if logged:
            logging.debug('[%s] -> %s' % (topic, data if data.__sizeof__() <= 1024 else ('%s bytes' % len(str(data)))))
        await self.client.publish(str(topic), data, retain=retain)

    async def hash_ctp_publish(self):
        try:
            await asyncio.sleep(1)
        except asyncio.CancelledError:
            pass
        else:
            await self.unsubscribe(mqtt.project.Hash(self.project_id))
            zip_data = utils.read_file(self.ctp_path, binary=True)

            with zipfile.ZipFile(self.ctp_path, 'r') as input_zip:
                def read_json(file):
                    with input_zip.open(file, "r") as fp:
                        return json.load(fp)

                tasks = (
                    asyncio.create_task(
                        self.publish(mqtt.project.ProjectFile.ctp(self.project_id), zip_data, logged=False)
                    ),
                    *(
                        asyncio.create_task(
                            self.publish(mqtt.project.ProjectFile(self.project_id, file), read_json(file), logged=False)
                        )
                        for file in input_zip.namelist()
                        if file not in (f'{mqtt.project.ProjectFile.MODELS}.json', f'{mqtt.project.ProjectFile.USERS}.json')
                    ),
                )

            try:
                await asyncio.gather(*tasks)
            except asyncio.CancelledError:
                logging.debug('Ignore cancelling publishing project')
            except Exception as ex:
                logging.exception(ex)

            await self.publish(mqtt.project.Hash(self.project_id), zlib.crc32(zip_data), logged=False)
            logging.debug('Hash and ctp published')

            logging.info('Run filling dashboard')
            os.system('sudo spread_utils_dashboard fill --all')
            logging.info('Filling done')

    async def project_publish(self):
        global start_time
        start_time = time.time()

        for topic in self.project:
            await self.publish(topic, self.project[topic], logged=False)

        if not self.project:
            logging.info("Project don't need publish")
        else:
            logging.info('Project published')
            # logging.debug(f"project published time: {time.time()-start_time}")

        self.project.clear()

    async def local_unsubscribe(self):
        try:
            await asyncio.sleep(1)
        except asyncio.CancelledError:
            pass
        else:
            # logging.debug(f"broker check time: {time.time()-start_time}")
            await self.unsubscribe(mqtt.project.Local.broadcast(self.project_id, '#'))
            await self.project_publish()
            self.BROKER_REVISION_STAGE = False
            await self.subscribe(mqtt.project.Set.broadcast(self.project_id, '#'))
            await self.subscribe(mqtt.project.Reset.broadcast(self.project_id, '#'))

    async def ctp_unsubscribe(self):
        try:
            await asyncio.sleep(1)
        except asyncio.CancelledError:
            pass
        else:
            await self.unsubscribe(f'{mqtt.project.ProjectTopic.protocol}/+/{mqtt.project.CTP}')

    async def project_init(self):
        """
            - Создаем словарь ctp в том виде, как он должен быть опубликован {топик_объект: значение}
            - project = ctp + изменения из файловой (storage)
        """
        global start_time
        errors = False
        start_time = time.time()
        self.ctp.clear()

        with zipfile.ZipFile(self.ctp_path, 'r') as input_zip:
            with input_zip.open(f"{mqtt.project.ProjectFile.HEADER}.json", "r") as file:
                header_file = json.load(file)['header']

            if self.project_id != header_file['project'].pop('id'):
                raise VerifyError(f'Init error: "{settings.PROJECT_ID}" in config file not corresponding of ctp project!')

            try:
                topic = mqtt.project.Header(self.project_id, 'Project')
                if self.verify_mode > 0:
                    passed = self.data_valid(topic, header_file['project'])

                if self.verify_mode < 2:
                    passed = True

                if passed:
                    self.ctp[topic] = header_file['project']
                else:
                    errors = True

            except Exception as ex:
                errors = True
                logging.debug(f"Init error: Wrong header ({ex})")

            try:
                topic = mqtt.project.Header(self.project_id, 'Firmware')
                if self.verify_mode > 0:
                    passed = self.data_valid(topic, header_file['firmware'])

                if self.verify_mode < 2:
                    passed = True

                if passed:
                    self.ctp[topic] = header_file['firmware']
                else:
                    errors = True

            except Exception as ex:
                errors = True
                logging.debug(f"Init error: Wrong header ({ex})")

            del header_file

            with input_zip.open(f"{mqtt.project.ProjectFile.SERVERS}.json", "r") as file:
                servers = json.load(file)['servers']

            self.SERVERS = []
            for obj in servers:
                if obj['id'] not in self.SERVERS:
                    self.SERVERS.append(obj['id'])
                else:
                    errors = True
                    logging.debug(f"Init error: Wrong server (id={obj['id']}, id not unique)")

            del servers

            with input_zip.open(f"{mqtt.project.ProjectFile.MANAGERS}.json", "r") as file:
                managers = json.load(file)['managers']

            self.MANAGERS = {}
            for obj in managers:
                obj_id = obj.pop('id')
                obj_type = obj.pop('type')
                if obj_id not in self.MANAGERS:
                    try:
                        topic = mqtt.project.Hardware(self.project_id, obj_type, obj_id)
                        if self.verify_mode > 0:
                            passed = self.data_valid(topic, obj)

                        if self.verify_mode < 2:
                            passed = True

                        if passed:
                            self.ctp[topic] = obj
                            self.MANAGERS[obj_id] = obj_type
                        else:
                            errors = True
                    except:
                        errors = True
                        logging.debug(f"Init error: Wrong manager (id={obj_id})")
                else:
                    errors = True
                    logging.debug(f"Init error: Wrong manager (id={obj_id}, id not unique)")

            del managers

            with input_zip.open(f"{mqtt.project.ProjectFile.PROVIDERS}.json", "r") as file:
                providers = json.load(file)['providers']

            self.PROVIDERS = {}
            for obj in providers:
                obj_id = obj.pop('id')
                obj_type = obj.pop('type')
                if obj_id not in self.PROVIDERS:
                    try:
                        topic = mqtt.project.Provider(self.project_id, self.MANAGERS[obj['managerID']], obj.pop('managerID'), obj_type, obj_id)
                        if self.verify_mode > 0:
                            passed = self.data_valid(topic, obj)

                        if self.verify_mode < 2:
                            passed = True

                        if passed:
                            self.ctp[topic] = obj
                            self.PROVIDERS[obj_id] = obj_type
                        else:
                            errors = True
                    except:
                        errors = True
                        logging.debug(f"Init error: Wrong provider (id={obj_id})")
                else:
                    errors = True
                    logging.debug(f"Init error: Wrong provider (id={obj_id}, id not unique)")

            del providers

            with input_zip.open(f"{mqtt.project.ProjectFile.ENGINERIES}.json", "r") as file:
                engineries = json.load(file)['engineries']

            self.ENGINERIES = {}
            for obj in engineries:
                obj_id = obj.pop('id')
                obj_type = obj.pop('type')
                if obj_id not in self.ENGINERIES:
                    try:
                        topic = mqtt.project.Equipment(self.project_id, obj_type, obj_id)
                        if self.verify_mode > 0:
                            passed = self.data_valid(topic, obj)

                        if self.verify_mode < 2:
                            passed = True

                        if passed:
                            self.ctp[topic] = obj
                            self.ENGINERIES[obj_id] = obj_type
                        else:
                            errors = True
                    except:
                        errors = True
                        logging.debug(f"Init error: Wrong enginery (id={obj_id})")
                else:
                    errors = True
                    logging.debug(f"Init error: Wrong enginery (id={obj_id}) (id not unique)")

            del engineries

            with input_zip.open(f"{mqtt.project.ProjectFile.LOCATIONS}.json", "r") as file:
                locations = json.load(file)['locations']

            self.LOCATIONS = []
            for obj in locations:
                obj_id = obj.pop('id')
                if obj_id not in self.LOCATIONS:
                    try:
                        topic = mqtt.project.Location(self.project_id, obj_id)
                        if self.verify_mode > 0:
                            passed = self.data_valid(topic, obj)

                        if self.verify_mode < 2:
                            passed = True

                        if passed:
                            self.ctp[topic] = obj
                            self.LOCATIONS.append(obj_id)
                        else:
                            errors = True
                    except:
                        errors = True
                        logging.debug(f"Init error: Wrong location (id={obj_id})")
                else:
                    errors = True
                    logging.debug(f"Init error: Wrong location (id={obj_id}, id not unique)")

            del locations

            with input_zip.open(f"{mqtt.project.ProjectFile.SUBGINERIES}.json", "r") as file:
                subgineries = json.load(file)['subgineries']

            SUBGINERIES = []
            for obj in subgineries:
                obj_id = obj.pop('id')
                # у Enginery и Subginery общая группа для проверки уникальности id
                if obj_id not in SUBGINERIES:
                    if obj_id not in self.ENGINERIES:
                        try:
                            topic = mqtt.project.Subginery(self.project_id, obj.pop('locationID'), obj.pop('type'), obj_id)
                            if self.verify_mode > 0:
                                passed = self.data_valid(topic, obj)

                            if self.verify_mode < 2:
                                passed = True

                            if passed:
                                self.ctp[topic] = obj
                                SUBGINERIES.append(obj_id)
                            else:
                                errors = True
                        except:
                            errors = True
                            logging.debug(f"Init error: Wrong subginery (id={obj_id})")
                    else:
                        errors = True
                        logging.debug(f"Init error: Wrong subginery (id={obj_id}, id not unique, enginery got the same id)")
                else:
                    errors = True
                    logging.debug(f"Init error: Wrong subginery (id={obj_id}, id not unique)")

            del subgineries, SUBGINERIES

            # users.json может отсутствовать в ctp-проекте!!!
            users_file = f"{mqtt.project.ProjectFile.USERS}.json"
            if users_file in input_zip.namelist():
                with input_zip.open(users_file, "r") as file:  # , encoding="utf-8" ???
                    users = json.load(file)['users']

                USERS = []
                for obj in users:
                    obj_id = obj.pop('id')
                    if obj_id not in USERS:
                        try:
                            topic = mqtt.project.User(self.project_id, obj_id)
                            if self.verify_mode > 0:
                                passed = self.data_valid(topic, obj)

                            if self.verify_mode < 2:
                                passed = True

                            if passed:
                                self.ctp[topic] = obj
                                USERS.append(obj_id)
                            else:
                                errors = True
                        except:
                            errors = True
                            logging.debug(f"Init error: Wrong user (id={obj_id})")
                    else:
                        errors = True
                        logging.debug(f"Init error: Wrong user (id={obj_id}, id not unique)")

                del users, USERS

        # logging.debug(f"ctp processing time: {time.time()-start_time}")

        await file_sys_storage_modify(self.client, settings.config[settings.LOCAL_PATH], self.project_id)

        # apply storage data to project
        self.CYCLES = []
        self.project = self.ctp.copy()
        for topic in self.storage.topics:
            if topic in self.project or isinstance(topic, (mqtt.project.Preset, mqtt.project.Cycle, mqtt.project.User)):
                try:
                    data = self.storage.load(topic)
                    if not data:
                        raise
                except:
                    errors = True
                    logging.debug(f'Init error: Entity in storage got wrong structure (topic="{topic}")')
                    continue

                # !!! В storage, Triggers хранятся в Cycle и trigger_id - str
                # !!! Если триггеры будут храниться отдельно, возможно, появиться проблема
                # !!! порядка считывания топиков: сначала Cycle, потом Triggers
                triggers = None
                if isinstance(topic, mqtt.project.Cycle) and 'triggers' in data:
                    triggers = data.pop('triggers')

                if self.verify_mode > 0:
                    passed = self.data_valid(topic, data)

                if self.verify_mode < 2:
                    passed = True

                if passed:
                    self.project[topic] = data
                else:
                    errors = True
                    logging.debug(f'Init error: Entity in storage got wrong structure (topic="{topic}")')

                if isinstance(topic, mqtt.project.Cycle):
                    self.CYCLES.append(topic.entity_id)
                    if triggers:
                        for tr_id in triggers:
                            try:
                                tr_topic = mqtt.project.CycleTrigger(topic.project_id, topic.entity_id, tr_id)
                                if self.verify_mode > 0:
                                    passed = self.data_valid(tr_topic, triggers[tr_id])

                                if self.verify_mode < 2:
                                    passed = True

                                if passed:
                                    self.project[tr_topic] = triggers[tr_id]
                                else:
                                    errors = True
                            except:
                                errors = True
                                logging.debug(f'Init error: Entity in storage got wrong structure (topic="{topic}")')
            else:
                errors = True
                logging.debug(f'Init error: Storage got wrong entity (topic="{topic}")')

        if errors:
            if self.verify_mode == 3:
                raise VerifyError('Project initialization error!')
            else:
                logging.info('Project initialization error!')

        logging.debug('Project ready to publish')

    def data_valid(self, topic: mqtt.project.Local, data) -> bool:
        """
            Возвращает True, если сущность (topic) из Local, может быть изменена и
            её данные (data) соответствуют ограничениям для этой сущности.
            Иначе возвращает False.
        """
        try:
            if isinstance(topic, mqtt.project.Equipment):  # Enginery
                enginery.EngineryValidator({'id': topic.entity_id, 'type': topic.entity_type, **data})

                # recipe ingredient id must be real provider_id and provider_type must be in Constraints
                if 'recipe' in data:
                    if 'ingredients' in data['recipe']:
                        for ingr in data['recipe']['ingredients']:
                            if ingr['id'] not in self.PROVIDERS:
                                raise ValueError('Wrong recipe ingredient id (not in project)')

                            if ingr['type'] in constr.ENG_CONSTR[topic.entity_type][data['recipe']['type']]:
                                ingr_type = ingr['type']
                            elif ingr['type'] + '[]' in constr.ENG_CONSTR[topic.entity_type][data['recipe']['type']]:
                                ingr_type = ingr['type'] + '[]'
                            else:
                                raise ValueError('Wrong recipe ingredient id type (not in constraints)')

                            if self.PROVIDERS[ingr['id']] not in constr.ENG_CONSTR[topic.entity_type][data['recipe']['type']][ingr_type]:
                                raise ValueError('Wrong recipe ingredient id type (not in constraints)')

            elif isinstance(topic, mqtt.project.Hardware):  # Manager
                manager.ManagerValidator({'id': topic.entity_id, 'type': topic.entity_type, **data})
                if data['serverID'] not in self.SERVERS:
                    raise ValueError('Wrong serverID (not in project)')

            elif isinstance(topic, mqtt.project.Provider):
                provider.ProviderValidator({'id': topic.entity_id, 'type': topic.entity_type, 'managerID': topic.manager_id, **data})

                # managerID must be real and manager_type must be in Constraints
                if topic.manager_id not in self.MANAGERS:
                    raise ValueError('Wrong managerID (not in project)')
                if topic.manager_type != self.MANAGERS[topic.manager_id]:
                    raise ValueError('Wrong managerID type')
                if topic.manager_type not in constr.MAN_PROV_CONSTR:
                    raise ValueError('Wrong managerID type (not in constraints)')
                if topic.entity_type not in constr.MAN_PROV_CONSTR[topic.manager_type]:
                    raise ValueError('Wrong provider type (not in constraints)')

            elif isinstance(topic, mqtt.project.Subginery):
                subginery.SubgineryValidator({'id': topic.entity_id, 'type': topic.entity_type, 'locationID': topic.location_id, **data})

                # engineries must be real and enginery_type must be in Constraints
                if 'engineries' in data:
                    for eng_id in data['engineries']:
                        if eng_id not in self.ENGINERIES:
                            raise ValueError('Wrong engineries (not in project)')
                        if self.ENGINERIES[eng_id] not in constr.SUBG_CONSTR[topic.entity_type]:
                            raise ValueError('Wrong engineries (not in constraints)')

            elif isinstance(topic, mqtt.project.Location):
                location.LocationValidator({'id': topic.entity_id, **data})

            elif isinstance(topic, mqtt.project.Preset):
                if topic.entity_type == ButtonPreset.__name__:
                    button_preset.ButtonPresetValidator({'id': topic.entity_id, **data})

                elif topic.entity_type == lom.profiles.DimmingProfile.__name__:
                    lom.profiles.GlobalDimmingProfile.deserialize({'id': topic.entity_id, **data})

                else:
                    raise ValueError('Wrong preset type')

            elif isinstance(topic, mqtt.project.Cycle):
                cycle.CycleValidator({'id': topic.entity_id, **data})

            elif isinstance(topic, mqtt.project.CycleTrigger):
                if topic.cycle_id not in self.CYCLES:
                    raise ValueError('Wrong cycle id')
                trigger.TriggerValidator({'id': topic.entity_id, **data})

            elif isinstance(topic, mqtt.project.User):
                user.UserValidator({'id': topic.entity_id, **data})
                if data['rootLocationID'] not in self.LOCATIONS:
                    raise ValueError('Wrong rootLocationID')

                if 'homeLocationID' in data and data['homeLocationID'] not in self.LOCATIONS:
                    raise ValueError('Wrong homeLocationID')

            elif isinstance(topic, mqtt.project.Header):
                if topic.entity_type == 'Project':
                    header_project.HeaderProjectValidator({'id': topic.project_id, **data})

                elif topic.entity_type == 'Firmware':
                    header_firmware.HeaderFirmwareValidator({**data})

                else:
                    raise ValueError('Wrong header type')
            else:
                return False

            return True
        except Exception as ex:
            logging.debug('Validation error: %s' % ex)
            return False

    async def set_topic(self, topic: mqtt.project.Local, data):
        if not self.data_valid(topic, data):
            raise ValueError('Set error: Invalid data type of topic: "%s"' % topic)

        if topic in self.ctp:
            storage_data = self.storage.load(topic)
            if data == self.ctp[topic]:
                if storage_data:
                    self.storage.save(topic, None)
                    await self.publish(topic, self.ctp[topic])

            elif data != storage_data:
                if (storage_data and data['key'] == storage_data['key']) or data['key'] == self.ctp[topic]['key']:
                    raise ValueError('Set error: New "key" needed')
                self.storage.save(topic, data)
                await self.publish(topic, data)
            # else:
            #     raise ValueError('Set error: No data change in topic: "%s"' % topic)

        elif isinstance(topic, mqtt.project.Preset):
            if data != self.storage.load(topic):
                self.storage.save(topic, data)
                await self.publish(topic, data)
            # else:
            #     raise ValueError('Set error: No data change in topic: "%s"' % topic)

        elif isinstance(topic, mqtt.project.Cycle):
            storage_data = self.storage.load(topic)
            if storage_data:
                triggers = storage_data.pop('triggers', None)
            else:
                triggers = None

            if data != storage_data:
                new_storage_data = data.copy()
                if triggers:
                    new_storage_data.update({'triggers': triggers})
                if topic.entity_id not in self.CYCLES:
                    self.CYCLES.append(topic.entity_id)
                self.storage.save(topic, new_storage_data)
                await self.publish(topic, data)
            # else:
            #     raise ValueError('Set error: No data change in topic: "%s"' % topic)

        elif isinstance(topic, mqtt.project.CycleTrigger):
            # В storage, Triggers хранятся в Cycle и trigger_id - str!!!
            cycle_topic = mqtt.project.Cycle(topic.project_id, topic.cycle_id)
            cycle_data = self.storage.load(cycle_topic)
            triggers = cycle_data.get('triggers')
            trig_id = str(topic.entity_id)

            if triggers and (trig_id in triggers) and data == triggers[trig_id]:
                return
                # raise ValueError('Set error: No data change in topic: "%s"' % topic)

            if not triggers:
                cycle_data['triggers'] = {}

            cycle_data['triggers'][trig_id] = data
            self.storage.save(cycle_topic, cycle_data)
            await self.publish(topic, data)
        else:
            raise ValueError('Set error: Invalid topic: "%s"' % topic)

    async def reset_topic(self, topic: mqtt.project.Local, data):
        if not isinstance(data, dict):
            raise ValueError('Reset error: Invalid data type of topic: "%s"' % topic)

        if topic in self.ctp:
            self.storage.save(topic, None)
            await self.publish(topic, self.ctp[topic])

        elif isinstance(topic, mqtt.project.Cycle):
            # При удалении Cycle, удаляем и Triggers из брокера
            storage_data = self.storage.load(topic)
            if storage_data:
                if 'triggers' in storage_data:
                    for tr_id in storage_data['triggers']:
                        await self.publish(mqtt.project.CycleTrigger(topic.project_id, topic.entity_id, tr_id), bytes())

                self.storage.save(topic, None)
                self.CYCLES.remove(topic.entity_id)

            await self.publish(topic, bytes())

        elif isinstance(topic, mqtt.project.CycleTrigger):
            # В storage, Triggers хранятся в Cycle и trigger_id - str!!!
            cycle_topic = mqtt.project.Cycle(topic.project_id, topic.cycle_id)
            cycle_data = self.storage.load(cycle_topic)
            if not cycle_data:
                await self.publish(topic, bytes())
            else:
                triggers = cycle_data.get('triggers')
                trig_id = str(topic.entity_id)

                if triggers and trig_id in triggers:
                    del cycle_data['triggers'][trig_id]
                    if not cycle_data['triggers']:
                        del cycle_data['triggers']
                    self.storage.save(cycle_topic, cycle_data)

                await self.publish(topic, bytes())
        else:
            self.storage.save(topic, None)
            await self.publish(topic, bytes())

        logging.debug(f'Reset topic: "%s"' % topic)

    async def listener(self):
        local_unsubscribe_task = None

        while True:
            logging.info('Projector started')
            logging.info(f'set verifyMode as "{self.verify_mode}"')
            try:
                async with Client(
                        hostname=settings.config[settings.BROKER].get(settings.HOST, '127.0.0.1'),
                        port=settings.config[settings.BROKER].get(settings.PORT, 1883),
                        username=settings.config[settings.BROKER].get(settings.LOGIN),
                        password=settings.config[settings.BROKER].get(settings.PASSWORD),
                        client_id='Projector'
                ) as client:
                    self.client = client

                    if local_unsubscribe_task:
                        local_unsubscribe_task.cancel()
                        local_unsubscribe_task = None

                    async with client.messages() as messages:
                        if self.BROKER_REVISION_STAGE:  # При старте сервиса True
                            try:
                                await self.project_init()
                            except VerifyError as ex:
                                logging.info(ex)
                                logging.info('Service stopped. To try start service anyway, ' + \
                                             'change "Verify mode" param in service config.')
                                break
                            except:
                                logging.exception(f'{sys.exc_info()[1]}')
                                logging.info('Service stopped.')
                                break

                            ctp_hash = zlib.crc32(utils.read_file(self.ctp_path, binary=True))
                            await self.subscribe(f'{mqtt.project.ProjectTopic.protocol}/+/{mqtt.project.CTP}')
                            await self.subscribe(mqtt.project.Hash(self.project_id))
                            await self.subscribe(mqtt.project.Local.broadcast(self.project_id, '#'))
                            asyncio.create_task(self.ctp_unsubscribe())
                            hash_ctp_publish_task = asyncio.create_task(self.hash_ctp_publish())
                            local_unsubscribe_task = asyncio.create_task(self.local_unsubscribe())
                            global start_time
                            start_time = time.time()
                        else:
                            await self.subscribe(mqtt.project.Set.broadcast(self.project_id, '#'))
                            await self.subscribe(mqtt.project.Reset.broadcast(self.project_id, '#'))

                        async for message in messages:
                            try:
                                if not message.topic.matches(f'{mqtt.project.ProjectTopic.protocol}/#'):
                                    continue

                                if not message.payload:
                                    continue

                                topic = mqtt.topic_of(message.topic.value)

                                if isinstance(topic, mqtt.project.ProjectFile):
                                    if topic.file_name == mqtt.project.ProjectFile.CTP:
                                        if topic.project_id != self.project_id:
                                            await self.publish(topic, bytes())
                                            await self.publish(mqtt.project.Hash(topic.project_id), bytes())
                                    continue

                                try:
                                    data = json.loads(message.payload)
                                except json.JSONDecodeError as ex:
                                    logging.error(f'JSONDecode error data of topic: "{message.topic.value}". {ex}')
                                    continue

                                if self.BROKER_REVISION_STAGE:
                                    if local_unsubscribe_task:
                                        local_unsubscribe_task.cancel()
                                        local_unsubscribe_task = None

                                    if isinstance(topic, mqtt.project.Hash):
                                        if data == ctp_hash:
                                            if hash_ctp_publish_task:
                                                hash_ctp_publish_task.cancel()
                                                hash_ctp_publish_task = None
                                                logging.info("Hash and ctp don't need publish")

                                            await self.unsubscribe(mqtt.project.Hash(self.project_id))

                                    elif issubclass(topic.__class__, mqtt.project.Local):
                                        if topic in self.project:
                                            if data == self.project[topic]:
                                                del self.project[topic]
                                        else:
                                            await self.publish(topic, bytes())

                                    local_unsubscribe_task = asyncio.create_task(self.local_unsubscribe())

                                elif isinstance(topic, mqtt.project.Set):
                                    topic = mqtt.topic_of(message.topic.value.replace(mqtt.project.SET, mqtt.project.LOCAL, 1))
                                    await self.set_topic(topic, data)

                                elif isinstance(topic, mqtt.project.Reset):
                                    topic = mqtt.topic_of(message.topic.value.replace(mqtt.project.RESET, mqtt.project.LOCAL, 1))
                                    await self.reset_topic(topic, data)

                            except json.JSONDecodeError as ex:
                                logging.error('%s' % ex)
                            except BaseException as ex:
                                logging.exception('%s' % ex)
            except MqttError as error:
                logging.error(error)
                await asyncio.sleep(settings.config.get(settings.RECONNECT_INTERVAL, 3))
            except (KeyboardInterrupt, asyncio.CancelledError):
                print('breaked')
                break
            except SystemExit as ex:
                if isinstance(ex, int) and ex == 100:
                    break
                else:
                    sys.exit(ex)
            except BaseException as ex:
                logging.exception('%s' % ex)


def run():
    projector = Projector()
    # signal.signal(signal.SIGHUP, lambda signum, frame: settings.config.parse())
    try:
        asyncio.run(projector.listener())
    except KeyboardInterrupt:
        pass

    return 100


if __name__ == '__main__':
    run()
