from unittest import TestCase

from spread_core.mqtt.spread.address import ProviderAddress

from spread_core.bam.engineries import climate
from spread_core.mqtt.spread.topic import State
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests.equipment.utils import TestEquipmentAdapter, TestEngineryRecipe


class _TestClimateFunits(TestCase):
    enginery_type = climate._Climate
    state_funits = ()
    set_funits = ()
    values = {}

    def setUp(self) -> None:
        self.adapter = TestEquipmentAdapter()
        self.recipe = TestEngineryRecipe()

        enginery = self.enginery_type(999)
        enginery.set_adapter(self.adapter)
        enginery.set_recipe('SomeType', self.recipe)

        self.enginery = enginery

    def _emulate_state(self, funit, value):
        self.enginery.on_state(State(ProviderAddress(1, 'Manager', 2, 'Provider', 3, funit)), Variable(value))

    def _check_state(self, funit, value, expect_state=True):
        self._emulate_state(funit, value)
        self._check_value(funit, value)

        if expect_state:
            self._check_publish(f'Spread/State/1/Equipment/{self.enginery_type.__name__}/999/{funit}', value)

        self._check_end_of_publishes()

    def _check_value(self, funit, expected_value):
        self.assertEqual(expected_value, self.enginery.values.get(funit))

    def _check_end_of_publishes(self):
        if self.adapter.queue:
            unexpected_publishes = f'Unexpected publishes ({len(self.adapter.queue)}):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.adapter.queue)
            self.fail(unexpected_publishes)

    def _check_publish(self, _topic, _value, _retain=True, _delay=0):
        topic, variable, retain, delay = self.adapter.queue.popleft()
        self.assertEqual(topic, _topic)
        self.assertEqual(variable.value, _value)
        self.assertEqual(retain, _retain)
        self.assertEqual(delay, _delay)

    def test_recipe_appointment(self):
        recipe = TestEngineryRecipe(1001)
        self.enginery.set_recipe('SomeString', recipe)
        self.assertIs(self.enginery.recipe, recipe)

    def test_address(self):
        self.assertEqual(self.enginery.get_address('SomeFunit'), f'1/Equipment/{self.enginery_type.__name__}/999/SomeFunit')

    def get_address_broadcast(self):
        self.assertEqual(self.enginery.get_address_broadcast('SomeFunit'), f'1/Equipment/{self.enginery_type.__name__}/+/SomeFunit')

    def test_recipe_selecting(self):
        funit, recipe = self.enginery.recipe_for_funit('SomeFunit')
        self.assertEqual(funit, 'SomeFunit')
        self.assertEqual(recipe, self.recipe)

    def test_state_handler(self):
        for funit in self.state_funits:
            for value in self.values[funit]:
                self._check_state(str(funit), value)
                self._check_state(str(funit), value, expect_state=False)

    def test_invalidation(self):
        self.enginery.values['Funit1'] = 1
        self.enginery.values['Funit2'] = 'Funit2'
        self.enginery.values['Funit3'] = [1, 2, 3]

        self.enginery.set_invalid()

        self._check_publish(f'Spread/State/1/Equipment/{self.enginery_type.__name__}/999/Funit1', None)
        self._check_publish(f'Spread/State/1/Equipment/{self.enginery_type.__name__}/999/Funit2', None)
        self._check_publish(f'Spread/State/1/Equipment/{self.enginery_type.__name__}/999/Funit3', None)

        self._check_end_of_publishes()

    def test_funits_set(self):
        for funit in self.set_funits:
            for value in self.values[funit]:
                funit(self.enginery, Variable(value)).set()
                self._check_publish(f'Spread/Set/Provider/1/{str(funit)}', value, False, 0)
                self._check_end_of_publishes()

    def test_funits_unsettable(self):
        for funit in self.state_funits:
            if funit not in self.set_funits:
                self.assertRaises(funit.NotImplemented, funit(self.enginery, Variable(100)).set)


class TestTemperatureSensor(_TestClimateFunits):
    enginery_type = climate.TemperatureSensor
    state_funits = (climate.Temperature, )
    set_funits = ()
    values = {climate.Temperature: [29815]}


class TestFan(_TestClimateFunits):
    enginery_type = climate.Fan
    state_funits = (climate.On, )
    set_funits = (climate.On, )
    values = {climate.On: [True, False]}


class TestTermoregulator(_TestClimateFunits):
    enginery_type = climate.Thermoregulator
    state_funits = (climate.Temperature, climate.TargetTemperature, climate.On, climate.Preset,
                    climate.Mode, climate.FanSpeedMode, climate.LouverMode)
    set_funits = (climate.TargetTemperature, climate.On, climate.Preset,
                  climate.Mode, climate.FanSpeedMode, climate.LouverMode)
    values = {
        climate.Temperature: [29815],
        climate.TargetTemperature: [29815],
        climate.On: [True, False],
        climate.Preset: ['Manual', 'Auto', 'Protection', 'Day', 'Night', 'Standby', 'Comfort', 'Timer'],
        climate.Mode: ['Manual', 'Auto', 'Cooling', 'Heating', 'Drying', 'Ventilation', 'HeatPlus', 'Stop'],
        climate.FanSpeedMode: ['Manual', 'Auto', 'High', 'Medium', 'Low', 'Top'],
        climate.LouverMode: ['Manual', 'Auto', 'Horizontal', 'Angle30', 'Angle45', 'Angle60', 'Vertical']
    }
