from unittest import TestCase

from spread_core.bam.engineries import Enginery

from spread_core.bam.engineries.climate import TemperatureSensor, Thermoregulator, Fan, HeatedFloor

from spread_core.bam.subgineries import climating as climating
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests.equipment.utils import TestEquipmentAdapter, TestEngineryRecipe


class TestClimate(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.adapter = TestEquipmentAdapter()

    def setUp(self) -> None:
        ENGS = [TemperatureSensor, Thermoregulator, Fan, HeatedFloor, Enginery]

        self.adapter.add_manager(2, "Manager")

        for i in range(15):
            e = ENGS[i//3](i)
            self.adapter.add_provider(TestEngineryRecipe(1000 + i))
            self.adapter.add_enginery(e, 1000+i, "Simple")

        self._sb = climating.Climate(99, 2, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
        self._sb.set_adapter(self.adapter)

    def _check_publish_state(self, funit, value, delay=0, params=None, attrs=None):
        self.__check_publish(f'Spread/State/1/Location/2/Climate/{funit}', value, _retain=True, _delay=delay, params=params, attrs=attrs)

    def _check_publish_command(self, p_id, funit, value, params=None):
        self.__check_publish(f'Spread/Set/Provider/{p_id}/{funit}', value, _retain=False, _delay=0, params=params)

    def _check_end_of_publishes(self):
        if self.adapter.queue:
            unexpected_publishes = f'Unexpected publishes({len(self.adapter.queue)} pcs):\n'
            unexpected_publishes += "\n".join(str(i) for i in self.adapter.queue)
            self.fail(unexpected_publishes)

    def __check_publish(self, _topic, _value, _retain=True, _delay=0, params=None, attrs=None):
        if len(self.adapter.queue) == 0:
            self.fail('Expected publish is absent: %s' % _topic)
        else:
            topic, variable, retain, delay = self.adapter.queue.popleft()
            self.assertEqual(_topic, topic)
            self.assertEqual(_value, variable.value)
            self.assertEqual(_retain, retain)
            self.assertEqual(_delay, delay)

            if params:
                for param_name in params:
                    try:
                        self.assertEqual(params[param_name], variable.params[param_name])
                    except KeyError:
                        self.fail('Expected variable parameter "%s" is absent' % param_name)

            if attrs:
                for attrs_name in attrs:
                    try:
                        self.assertEqual(attrs[attrs_name], getattr(variable, attrs_name))
                    except KeyError:
                        self.fail('Expected variable attribute "%s" is absent' % attrs_name)

    @property
    def sb(self) -> climating.Climate:
        return self._sb

    def test_init(self):
        self.assertEqual(1, self.sb.project_id)
        self.assertEqual(2, self.sb.location_id)
        self.assertEqual(99, self.sb.id)
        self.assertEqual(self.sb._enginery_ids, [e.id for e in self.sb.engineries])

        for e in self.sb.engineries:
            assert isinstance(e, Enginery)
            self.assertIn(self.sb, e._subgineries)

    def test_address(self):
        self.assertEqual('1/Location/2/Climate/SomeFunit', self.sb.get_address("SomeFunit"))

    def test_state_publishing(self):
        self.sb.publish_state('SomeFunit', 'SomeValue')
        self._check_publish_state('SomeFunit', 'SomeValue')
        self._check_end_of_publishes()

    def test_invalidation(self):
        self.sb.values['SomeFunit'] = 2
        self.sb.set_invalid()

        self._check_publish_state('SomeFunit', None)
        self._check_end_of_publishes()

    def _set_enginery_value(self, enginery_id, eng_funit_type, value, expected_unit=None, expected_value=None, **attrs):
        e = self.adapter._entities[enginery_id]
        assert isinstance(e, Enginery)
        e.values[eng_funit_type] = value
        self.sb.on_level_changed(e, eng_funit_type)
        if expected_unit:
            self._check_publish_state(expected_unit, expected_value, attrs=attrs)

        self._check_end_of_publishes()

    def test_temperature_changing(self):
        self._set_enginery_value(0, 'Temperature', 29815, 'Temperature', 29815, invalid=True, inconsistent=False)
        self._set_enginery_value(1, 'Temperature', 29815)
        self._set_enginery_value(2, 'Temperature', 29815, 'Temperature', 29815, invalid=False, inconsistent=False)

        self._set_enginery_value(5, 'Temperature', 29415)
        self._set_enginery_value(8, 'Temperature', 29415)
        self._set_enginery_value(11, 'Temperature', 29415)

        self._set_enginery_value(2, 'Temperature', 29415, 'Temperature', 29681, invalid=False, inconsistent=True)
        self._set_enginery_value(0, 'Temperature', None, 'Temperature', 29615, invalid=True, inconsistent=True)
        self._set_enginery_value(1, 'Temperature', None, 'Temperature', 29415, invalid=True, inconsistent=False)
        self._set_enginery_value(2, 'Temperature', None, 'Temperature', None, invalid=True, inconsistent=False)

    def test_fan_changing(self):
        self._set_enginery_value(6, 'On', True, 'FansOn', True, invalid=True, inconsistent=False)
        self._set_enginery_value(7, 'On', True)
        self._set_enginery_value(8, 'On', True, 'FansOn', True, invalid=False, inconsistent=False)

        self._set_enginery_value(2, 'On', False)
        self._set_enginery_value(5, 'On', False, 'ThermoregulatorsOn', False, invalid=True, inconsistent=False)
        self._set_enginery_value(11, 'On', False, 'HeatedFloorsOn', False, invalid=True, inconsistent=False)

        self._set_enginery_value(8, 'On', False, 'FansOn', None, invalid=False, inconsistent=True)
        self._set_enginery_value(6, 'On', None, 'FansOn', None, invalid=True, inconsistent=True)
        self._set_enginery_value(7, 'On', None, 'FansOn', False, invalid=True, inconsistent=False)
        self._set_enginery_value(8, 'On', None, 'FansOn', None, invalid=True, inconsistent=False)

    def test_heatedfloor_changing(self):
        self._set_enginery_value(9, 'On', True, 'HeatedFloorsOn', True, invalid=True, inconsistent=False)
        self._set_enginery_value(10, 'On', True)
        self._set_enginery_value(11, 'On', True, 'HeatedFloorsOn', True, invalid=False, inconsistent=False)

        self._set_enginery_value(2, 'On', False)
        self._set_enginery_value(5, 'On', False, 'ThermoregulatorsOn', False, invalid=True, inconsistent=False)
        self._set_enginery_value(8, 'On', False, 'FansOn', False, invalid=True, inconsistent=False)

        self._set_enginery_value(11, 'On', False, 'HeatedFloorsOn', None, invalid=False, inconsistent=True)
        self._set_enginery_value(9, 'On', None, 'HeatedFloorsOn', None, invalid=True, inconsistent=True)
        self._set_enginery_value(10, 'On', None, 'HeatedFloorsOn', False, invalid=True, inconsistent=False)
        self._set_enginery_value(11, 'On', None, 'HeatedFloorsOn', None, invalid=True, inconsistent=False)

    def test_thermoregulator_changing(self):
        self._set_enginery_value(3, 'On', True, 'ThermoregulatorsOn', True, invalid=True, inconsistent=False)
        self._set_enginery_value(4, 'On', True)
        self._set_enginery_value(5, 'On', True, 'ThermoregulatorsOn', True, invalid=False, inconsistent=False)

        self._set_enginery_value(8, 'On', False, 'FansOn', False, invalid=True, inconsistent=False)
        self._set_enginery_value(11, 'On', False, 'HeatedFloorsOn', False, invalid=True, inconsistent=False)

        self._set_enginery_value(5, 'On', False, 'ThermoregulatorsOn', None, invalid=False, inconsistent=True)

        self._set_enginery_value(3, 'On', None, 'ThermoregulatorsOn', None, invalid=True, inconsistent=True)
        self._set_enginery_value(4, 'On', None, 'ThermoregulatorsOn', False, invalid=True, inconsistent=False)
        self._set_enginery_value(5, 'On', None, 'ThermoregulatorsOn', None, invalid=True, inconsistent=False)

        self._set_enginery_value(3, 'TargetTemperature', 29815, 'TargetTemperature', 29815, invalid=True, inconsistent=False)
        self._set_enginery_value(4, 'TargetTemperature', 29815)
        self._set_enginery_value(5, 'TargetTemperature', 29815, 'TargetTemperature', 29815, invalid=False, inconsistent=False)

        self._set_enginery_value(5, 'TargetTemperature', 29715, 'TargetTemperature', None, invalid=False, inconsistent=True)
        self._set_enginery_value(3, 'TargetTemperature', None, 'TargetTemperature', None, invalid=True, inconsistent=True)
        self._set_enginery_value(4, 'TargetTemperature', None, 'TargetTemperature', 29715, invalid=True, inconsistent=False)
        self._set_enginery_value(5, 'TargetTemperature', None, 'TargetTemperature', None, invalid=True, inconsistent=False)

        self._set_enginery_value(3, 'Preset', 'Manual', 'Preset', 'Manual', invalid=True, inconsistent=False)
        self._set_enginery_value(4, 'Preset', 'Manual')
        self._set_enginery_value(5, 'Preset', 'Manual', 'Preset', 'Manual', invalid=False, inconsistent=False)

        self._set_enginery_value(5, 'Preset', 'Auto', 'Preset', None, invalid=False, inconsistent=True)

        self._set_enginery_value(3, 'Preset', None, 'Preset', None, invalid=True, inconsistent=True)
        self._set_enginery_value(4, 'Preset', None, 'Preset', 'Auto', invalid=True, inconsistent=False)
        self._set_enginery_value(5, 'Preset', None, 'Preset', None, invalid=True, inconsistent=False)

        self._set_enginery_value(3, 'Mode', 'Manual', 'Mode', 'Manual', invalid=True, inconsistent=False)
        self._set_enginery_value(4, 'Mode', 'Manual')
        self._set_enginery_value(5, 'Mode', 'Manual', 'Mode', 'Manual', invalid=False, inconsistent=False)

        self._set_enginery_value(5, 'Mode', 'Auto', 'Mode', None, invalid=False, inconsistent=True)

        self._set_enginery_value(3, 'Mode', None, 'Mode', None, invalid=True, inconsistent=True)
        self._set_enginery_value(4, 'Mode', None, 'Mode', 'Auto', invalid=True, inconsistent=False)
        self._set_enginery_value(5, 'Mode', None, 'Mode', None, invalid=True, inconsistent=False)

        self._set_enginery_value(3, 'FanSpeedMode', 'Manual', 'FanSpeedMode', 'Manual', invalid=True, inconsistent=False)
        self._set_enginery_value(4, 'FanSpeedMode', 'Manual')
        self._set_enginery_value(5, 'FanSpeedMode', 'Manual', 'FanSpeedMode', 'Manual', invalid=False, inconsistent=False)

        self._set_enginery_value(5, 'FanSpeedMode', 'Auto', 'FanSpeedMode', None, invalid=False, inconsistent=True)

        self._set_enginery_value(3, 'FanSpeedMode', None, 'FanSpeedMode', None, invalid=True, inconsistent=True)
        self._set_enginery_value(4, 'FanSpeedMode', None, 'FanSpeedMode', 'Auto', invalid=True, inconsistent=False)
        self._set_enginery_value(5, 'FanSpeedMode', None, 'FanSpeedMode', None, invalid=True, inconsistent=False)

        self._set_enginery_value(3, 'LouverMode', 'Manual', 'LouverMode', 'Manual', invalid=True, inconsistent=False)
        self._set_enginery_value(4, 'LouverMode', 'Manual')
        self._set_enginery_value(5, 'LouverMode', 'Manual', 'LouverMode', 'Manual', invalid=False, inconsistent=False)

        self._set_enginery_value(5, 'LouverMode', 'Auto', 'LouverMode', None, invalid=False, inconsistent=True)

        self._set_enginery_value(3, 'LouverMode', None, 'LouverMode', None, invalid=True, inconsistent=True)
        self._set_enginery_value(4, 'LouverMode', None, 'LouverMode', 'Auto', invalid=True, inconsistent=False)
        self._set_enginery_value(5, 'LouverMode', None, 'LouverMode', None, invalid=True, inconsistent=False)

    def test_inappropriate_changing(self):
        self._set_enginery_value(0, 'On', True)
        self._set_enginery_value(0, 'TargetTemperature', 29815)
        self._set_enginery_value(0, 'Preset', 'Auto')
        self._set_enginery_value(0, 'Mode', 'Auto')
        self._set_enginery_value(0, 'FanSpeedMode', 'Auto')
        self._set_enginery_value(0, 'LouverMode', 'Auto')

        self._set_enginery_value(3, 'Temperature', 29815)

        self._set_enginery_value(6, 'Temperature', True)
        self._set_enginery_value(6, 'TargetTemperature', 29815)
        self._set_enginery_value(6, 'Preset', 'Auto')
        self._set_enginery_value(6, 'Mode', 'Auto')
        self._set_enginery_value(6, 'FanSpeedMode', 'Auto')
        self._set_enginery_value(6, 'LouverMode', 'Auto')

        self._set_enginery_value(9, 'Temperature', True)
        self._set_enginery_value(9, 'TargetTemperature', 29815)
        self._set_enginery_value(9, 'Preset', 'Auto')
        self._set_enginery_value(9, 'Mode', 'Auto')
        self._set_enginery_value(9, 'FanSpeedMode', 'Auto')
        self._set_enginery_value(9, 'LouverMode', 'Auto')

    def _check_command(self, funit, provider_ids, values, command):
        for value in values:
            funit(self.sb, Variable(value)).set()
            for provider_id in provider_ids:
                self._check_publish_command(provider_id, command, value)

            self._check_end_of_publishes()

    def test_temperature_setting(self):
        self.assertRaises(climating.Temperature.NotImplemented, climating.Temperature(self.sb, Variable(100)).set)

    def test_thermoregulators_on_setting(self):
        self._check_command(climating.ThermoregulatorsOn, [1000+i for i in range(3, 6)], (True, False), "On")

    def test_fans_on_setting(self):
        self._check_command(climating.FansOn, [1000+i for i in range(6, 9)], (True, False), "On")

    def test_heatedfloors_on_setting(self):
        self._check_command(climating.HeatedFloorsOn, [1000+i for i in range(9, 12)], (True, False), "On")

    def test_targettemperature_setting(self):
        self._check_command(climating.TargetTemperature, [1000+i for i in range(3, 6)], (29815, ), "TargetTemperature")

    def test_preset_setting(self):
        values = ('Manual', 'Auto', 'Protection', 'Day', 'Night', 'Standby', 'Comfort', 'Timer')
        self._check_command(climating.Preset, [1000+i for i in range(3, 6)], values, "Preset")

    def test_mode_setting(self):
        values = ('Manual', 'Auto', 'Cooling', 'Heating', 'Drying', 'Ventilation', 'HeatPlus', 'Stop')
        self._check_command(climating.Mode, [1000+i for i in range(3, 6)], values, "Mode")

    def test_fanspeedmode_setting(self):
        values = ('Manual', 'Auto', 'High', 'Medium', 'Low', 'Top')
        self._check_command(climating.FanSpeedMode, [1000+i for i in range(3, 6)], values, "FanSpeedMode")

    def test_louvermode_setting(self):
        values = ('Manual', 'Auto', 'Horizontal', 'Angle30', 'Angle45', 'Angle60', 'Vertical')
        self._check_command(climating.LouverMode, [1000+i for i in range(3, 6)], values, "LouverMode")
