from unittest import mock

import spread_dali as dali
from spread_dali.gear import general as gear
from .test_dali_dimmer import Tuner, TestDaliLight


class TestTuner(TestDaliLight):
    def setUp(self) -> None:
        super(TestTuner, self).setUp()
        self.set_value('MinLevelRaw', 100)
        self.set_value('MaxLevelRaw', 200)
        self.tuner = Tuner(self.provider)
        self.tuner._cmd = object()

    def emulate_command(self, command):
        self.tuner.check_state(command(dali.address.Short(63)))


@mock.patch.object(Tuner, Tuner.stop.__name__)
@mock.patch.object(Tuner, Tuner._stop_timer.__name__)
@mock.patch.object(Tuner, Tuner._start_timer.__name__)
class TestTunerSteps(TestTuner):
    def setUp(self) -> None:
        super(TestTunerSteps, self).setUp()
        self.set_value('MinLevelRaw', 100)
        self.set_value('MaxLevelRaw', 200)
        self.tuner = Tuner(self.provider)
        self.tuner._cmd = object()

    def emulate_command(self, command, has_command='matches'):
        cmd_inst = command(dali.address.Short(63))

        if has_command == 'matches':
            has_command = cmd_inst

        self.tuner._cmd = has_command
        self.tuner.check_state(cmd_inst)

    def test_uncommanded(self, start, stop, destroy):
        self.emulate_command(gear.Off, has_command=None)
        stop.assert_called_once()
        start.assert_not_called()
        destroy.assert_not_called()

    def _check_last_time(self, command, start, stop, destroy):
        self.set_value('CurrentLevelRaw', 150)
        self.tuner.count = 254
        self.emulate_command(command)
        destroy.assert_not_called()
        start.assert_not_called()
        stop.assert_not_called()

    def test_last_time_on_up(self, start, stop, destroy):
        self._check_last_time(gear.Up, start, stop, destroy)

    def test_last_time_on_step_up(self, start, stop, destroy):
        self._check_last_time(gear.StepUp, start, stop, destroy)

    def test_last_time_on_down(self, start, stop, destroy):
        self._check_last_time(gear.Down, start, stop, destroy)

    def test_last_time_on_step_down(self, start, stop, destroy):
        self._check_last_time(gear.StepDown, start, stop, destroy)

    def test_last_time_on_on_and_step_up(self, start, stop, destroy):
        self._check_last_time(gear.OnAndStepUp, start, stop, destroy)

    def test_last_time_on_step_down_and_off(self, start, stop, destroy):
        self._check_last_time(gear.StepDownAndOff, start, stop, destroy)

    def _check_max(self, command, levels: [int, int], destroy):
        self.set_value('CurrentLevelRaw', levels[0])
        self.emulate_command(command)
        destroy.assert_not_called()

        self.set_value('CurrentLevelRaw', levels[1])
        self.emulate_command(command)
        destroy.assert_called_once()

    def test_on_and_step_up(self, start, stop, destroy):
        self._check_max(gear.OnAndStepUp, [199, 200], destroy)

    def test_step_down_and_off(self, start, stop, destroy):
        self._check_max(gear.StepDownAndOff, [1, 0], destroy)

    def test_up(self, start, stop, destroy):
        self._check_max(gear.Up, [199, 200], destroy)

    def test_down(self, start, stop, destroy):
        self._check_max(gear.Down, [101, 100], destroy)

    def test_step_up(self, start, stop, destroy):
        self._check_max(gear.StepUp, [199, 200], destroy)

    def test_step_down(self, start, stop, destroy):
        self._check_max(gear.StepDown, [101, 100], destroy)


class TestTunerStopStart(TestTuner):
    def test_stop(self):
        self.tuner.set_direction(gear.StepDown, 99)
        self.tuner.stop()
        self.assertIsNone(self.tuner._cmd)
        self.assertIsNone(self.provider.tuner)
        self.assertEqual(0, self.tuner._delay)
        self.assertIsNone(self.tuner.timer)

    @mock.patch.object(Tuner, Tuner._start_timer.__name__)
    def test_start(self, timer_mock):
        self.tuner.set_direction(gear.StepDown, 99)
        self.assertIsInstance(self.tuner._cmd, gear.StepDown)
        self.assertEqual(99, self.tuner.speed)
        timer_mock.assert_called_once_with(2)
