import unittest
import unittest.mock as mock
from asyncio import coroutine

import spread_core.bam.managers as m_lib
import spread_dali as dali
from spread_core.bam.dali.managers import Dali
from spread_core.bam.dali.providers.providers import DaliProvider, Funit, Gtin, Serial, GtinOem, SerialOem, FirmwareVersion, HardwareVersion
from spread_core.bam.dali.tuner import TunerLimited
from spread_core.mqtt.spread.variable import Variable
from spread_core.tools.signature import CommandSignature
from spread_core.tests.hardware.utils import MockDali, MockBus


@coroutine
def response_generator():
    x = 1
    while True:
        yield dali.command.Response(dali.frame.BackwardFrame(x))
        x += 1


class TestDaliProvider(unittest.TestCase):
    Provider = DaliProvider

    @classmethod
    def setUpClass(cls) -> None:
        import logging
        logging.getLogger().setLevel(logging.WARNING)

        TunerLimited.UP = dali.gear.general.StepUp
        TunerLimited.DOWN = dali.gear.general.StepDown

        m_lib.config.args = [None]

    def setUp(self) -> None:
        self.bus = MockBus.of_dali()
        self.manager = MockDali(self.bus)

        self.setUpProvider()

    def setUpProvider(self):
        provider = self.Provider(999, dali.address.Short(63))
        provider.set_manager(self.manager)
        self._provider = provider

    @property
    def provider(self) -> DaliProvider:
        return self._provider

    def _resp(self, value):
        return dali.command.Response(dali.frame.BackwardFrame(value))

    def _make_update(self, funit, value, session=None, key=None, sig=None):
        if sig is None:
            if session:
                sig = CommandSignature(funit, key=key, session_id=session)
            else:
                sig = None

        self.provider.on_update(funit, value, sig=sig)

    def _check_value(self, funit, expected_value):
        self.assertEqual(expected_value, self.provider.values[funit])

    def _check_state(self, funit, value, check_value=True, **attrs):
        if check_value:
            self._check_value(funit, value)
        self.manager.check_publish(self, f'Spread/State/1/Hardware/MockDali/2/{self.provider.__class__.__name__}/999/{funit}', value, attrs=attrs)

    def _check_reply(self, session, key, value):
        self.manager.check_publish(self, f'Spread/Reply/1/{session}', value, _retain=False, attrs={'key': key})

    def _check_end_of_publishes(self):
        self.manager.check_end_of_publishes(self)

    def _set_dtr(self, index, value):
        self.manager.set_dtr_value(1, index, value)

    def _emulate_request_response(self, funit, value, *args):
        self.provider.on_request_answer(funit, dali.command.Response(dali.frame.BackwardFrame(value)), *args)

    def _emulate_external(self, command):
        self.provider.on_command_sended(command)

    def set_value(self, funit, value):
        if value is None:
            self.provider.values.pop(funit, None)
        else:
            self.provider.values[funit] = value


class TestUpdateHandling(TestDaliProvider):
    def test_easy(self):
        self._make_update('SomeUnit', 852)
        self._check_state('SomeUnit', 852)

        self._make_update('SomeUnit', 852)
        self._check_end_of_publishes()

    @mock.patch.object(Dali, 'add_to_group')
    @mock.patch.object(Dali, 'remove_from_group')
    def test_groups(self, remove_mock, add_mock):
        value = [
            {'index': 0, 'value': True},
            {'index': 7, 'value': False},
            {'index': 15, 'value': True},
            {'index': 22, 'value': False},
            {'index': 30, 'value': True},
        ]

        self._make_update('Groups', value)

        self.assertEqual(3, add_mock.call_count)
        self.assertEqual(2, remove_mock.call_count)

        add_mock.assert_has_calls([
            mock.call(self.provider, 0),
            mock.call(self.provider, 15),
            mock.call(self.provider, 30)
        ])

        remove_mock.assert_has_calls([
            mock.call(self.provider, 7),
            mock.call(self.provider, 22)
        ])

        self._check_state('Groups', value)
        self._check_end_of_publishes()


@mock.patch.object(DaliProvider, 'restore_states')
@mock.patch.object(DaliProvider, 'remove_states')
class TestUpdateValidation(TestDaliProvider):
    @property
    def provider_status(self):
        return self.provider._status

    @property
    def is_provider_valid(self):
        return self.provider.is_valid

    def _set_status(self, has_status, validity):
        self.provider._status = has_status
        self._make_update('Validity', validity)

    def test_validation_0_0(self, remover, restorer):
        self._set_status('invalid', False)
        self.assertEqual('invalid', self.provider_status)
        self.assertFalse(self.is_provider_valid)
        remover.assert_not_called()
        restorer.assert_not_called()

    def test_validation_1_0(self, remover, restorer):
        self._set_status('suspicious', False)
        self.assertEqual('invalid', self.provider_status)
        self.assertFalse(self.is_provider_valid)
        restorer.assert_not_called()
        remover.assert_called_once()

    def test_validation_2_0(self, remover, restorer):
        self._set_status('valid', False)
        self.assertEqual('suspicious', self.provider_status)
        self.assertTrue(self.is_provider_valid)
        remover.assert_not_called()
        restorer.assert_not_called()

    def test_validation_0_1(self, remover, restorer):
        self._set_status('invalid', True)
        self.assertEqual('valid', self.provider_status)
        self.assertTrue(self.is_provider_valid)
        remover.assert_not_called()
        restorer.assert_called_once()

    def test_validation_1_1(self, remover, restorer):
        self._set_status('suspicious', True)
        self.assertEqual('valid', self.provider_status)
        self.assertTrue(self.is_provider_valid)
        remover.assert_not_called()
        restorer.assert_not_called()

    def test_validation_2_1(self, remover, restorer):
        self._set_status('valid', True)
        self.assertEqual('valid', self.provider_status)
        self.assertTrue(self.is_provider_valid)
        remover.assert_not_called()
        restorer.assert_not_called()


class TestUpdateSigning(TestDaliProvider):
    def test_session(self):
        self._make_update('SomeFunit', 852, session='some_session')

        self._check_state('SomeFunit', 852)
        self._check_reply('some_session', None, 852)
        self._check_end_of_publishes()

    def test_key(self):
        self._make_update('SomeFunit', 852, key='some_key')

        self._check_state('SomeFunit', 852)
        self._check_end_of_publishes()

    def test_session_key(self):
        self._make_update('SomeFunit', 852, session='some_session', key='some_key')

        self._check_state('SomeFunit', 852)
        self._check_reply('some_session', 'some_key', 852)
        self._check_end_of_publishes()


@mock.patch.object(DaliProvider, '_parse_groups')
class TestUpdatePartsOfGroups(TestDaliProvider):
    def test_0_7(self, part_mock):
        self._make_update('Groups0-7', 111)
        part_mock.assert_called_once_with(0, 111, sig=None)

    def test_8_15(self, part_mock):
        self._make_update('Groups8-15', 222)
        part_mock.assert_called_once_with(8, 222, sig=None)

    def test_16_23(self, part_mock):
        sig = object()
        self._make_update('Groups16-23', 333, sig=sig)
        part_mock.assert_called_once_with(16, 333, sig=sig)

    def test_24_31(self, part_mock):
        self._make_update('Groups24-31', 444, session='qwe')
        part_mock.assert_called_once()
        self.assertEqual(24, part_mock.call_args[0][0])
        self.assertEqual(444, part_mock.call_args[0][1])
        self.assertEqual('qwe', part_mock.call_args[1]['sig'].session_id)


@mock.patch.object(DaliProvider, 'update_recursive')
class TestGroupParsing(TestDaliProvider):
    def test_bad_index(self, update_mock):
        with self.assertRaises(AssertionError) as ex:
            self.provider._parse_groups(1, 0)

        self.assertEqual('Incorrect first index', ex.exception.args[0])
        self.assertFalse(update_mock.called)

    def test_1_1(self, update_mock):
        sig = object()

        value = [{'index': i, 'value': False} for i in range(16)]
        value[1]['value'] = True
        value[4]['value'] = True
        value[6]['value'] = True

        self.provider._dali_type = 1
        self.provider._parse_groups(0, int('01010010', 2), sig=sig)
        update_mock.assert_called_once_with('Groups', value, sig=sig)

    def test_1_2(self, update_mock):
        value = [{'index': i, 'value': False} for i in range(16)]
        value[8]['value'] = True
        value[11]['value'] = True
        value[15]['value'] = True

        self.provider._dali_type = 1
        self.provider._parse_groups(8, int('10001001', 2))
        update_mock.assert_called_once_with('Groups', value, sig=None)

    def test_2_1(self, update_mock):
        sig = object()

        value = [{'index': i, 'value': False} for i in range(32)]
        value[1]['value'] = True
        value[4]['value'] = True
        value[6]['value'] = True

        self.provider._dali_type = 2
        self.provider._parse_groups(0, int('01010010', 2), sig=sig)
        update_mock.assert_called_once_with('Groups', value, sig=sig)

    def test_2_4(self, update_mock):
        value = [{'index': i, 'value': False} for i in range(32)]
        value[24]['value'] = True
        value[27]['value'] = True
        value[31]['value'] = True

        self.provider._dali_type = 2
        self.provider._parse_groups(24, int('10001001', 2))
        update_mock.assert_called_once_with('Groups', value, sig=None)


class TestMemoryPartParsing(TestDaliProvider):
    def setUp(self) -> None:
        super(TestMemoryPartParsing, self).setUp()

    def _test_number(self, funit: str, bank_id: int, start_index: int, end_index: int):
        responses = response_generator()
        expected_value = ''
        for location_id in range(start_index, end_index):
            self.provider.on_request_answer('MemoryPart', next(responses), location_id, bank_id)
            expected_value += f'0{location_id - start_index + 1}'
            self._check_state(funit, expected_value.ljust(2 * (end_index - start_index), '.'))

        self._check_end_of_publishes()

    def test_unknown_bank_id(self):
        with self.assertRaises(ValueError) as ex:
            self.provider.on_request_answer('MemoryPart', next(response_generator()), 0x3, 99)

        self.assertEqual('Unknown memory part of [99, 3]', ex.exception.args[0])

    def test_unknown_location_id(self):
        with self.assertRaises(ValueError) as ex:
            self.provider.on_request_answer('MemoryPart', next(response_generator()), 99, 1)

        self.assertEqual('Unknown memory part of [1, 99]', ex.exception.args[0])

    def test_gtin(self):
        self.provider.on_request_answer('MemoryPart', self._resp(1), 0x3, 0)
        self._check_state('Gtin', '01..........')

        self.provider.on_request_answer('MemoryPart', self._resp(2), 0x8, 0)
        self._check_state('Gtin', '01........02')

        self.provider.on_request_answer('MemoryPart', self._resp(3), 0x5, 0)
        self._check_state('Gtin', '01..03....02')

    def test_serial(self):
        self.provider.on_request_answer('MemoryPart', self._resp(1), 0xb, 0)
        self._check_state('Serial', '01..............')

        self.provider.on_request_answer('MemoryPart', self._resp(2), 0x12, 0)
        self._check_state('Serial', '01............02')

        self.provider.on_request_answer('MemoryPart', self._resp(3), 0xf, 0)
        self._check_state('Serial', '01......03....02')
        self._check_end_of_publishes()

    def test_gtin_oem(self):
        self.provider.on_request_answer('MemoryPart', self._resp(1), 0x3, 1)
        self._check_state('GtinOem', '01..........')

        self.provider.on_request_answer('MemoryPart', self._resp(2), 0x8, 1)
        self._check_state('GtinOem', '01........02')

        self.provider.on_request_answer('MemoryPart', self._resp(3), 0x5, 1)
        self._check_state('GtinOem', '01..03....02')
        self._check_end_of_publishes()

    def test_serial_oem(self):
        self.provider.on_request_answer('MemoryPart', self._resp(1), 0x9, 1)
        self._check_state('SerialOem', '01..............')

        self.provider.on_request_answer('MemoryPart', self._resp(2), 0x10, 1)
        self._check_state('SerialOem', '01............02')

        self.provider.on_request_answer('MemoryPart', self._resp(3), 0xd, 1)
        self._check_state('SerialOem', '01......03....02')
        self._check_end_of_publishes()

    def test_firmware_version(self):
        self.provider.on_request_answer('MemoryPart', self._resp(1), 0xa, 0)
        self._check_state('FirmwareVersion', '.1')

        self.provider.on_request_answer('MemoryPart', self._resp(2), 0x9, 0)
        self._check_state('FirmwareVersion', '2.1')

        self._check_end_of_publishes()

    def test_hardware_version(self):
        self.provider.on_request_answer('MemoryPart', self._resp(1), 0x14, 0)
        self._check_state('HardwareVersion', '.1')

        self.provider.on_request_answer('MemoryPart', self._resp(2), 0x13, 0)
        self._check_state('HardwareVersion', '2.1')

        self._check_end_of_publishes()


class TestFunitGtin(TestDaliProvider):
    call_args = [0, 0x3, 0x9]
    FunitClass = Gtin
    expected_result = '0001020304FF'

    def setUp(self) -> None:
        super(TestFunitGtin, self).setUp()
        self.funit = self.FunitClass(self.provider, Variable(None))

    def test_set(self):
        self.assertRaises(Funit.NotImplemented, self.funit.set)

    @mock.patch.object(DaliProvider, 'read_memory_location')
    @mock.patch.object(DaliProvider, 'on_update')
    def test_get(self, update_mock, memory_mock):
        memory_mock.return_value = b'\x00\x01\x02\x03\x04\xff'
        self.funit.get()
        memory_mock.assert_called_once_with(*self.call_args)
        update_mock.assert_called_once_with(self.FunitClass.__name__, self.expected_result, sig=None)


class TestFunitSerial(TestFunitGtin):
    FunitClass = Serial
    call_args = [0, 0xb, 0x13]


class TestFunitGtinOem(TestFunitGtin):
    FunitClass = GtinOem
    call_args = [1, 0x3, 0x9]


class TestFunitSerialOem(TestFunitGtin):
    FunitClass = SerialOem
    call_args = [1, 0x9, 0x11]


class TestFunitFirmwareVersion(TestFunitGtin):
    FunitClass = FirmwareVersion
    call_args = [0, 0x9, 0xb]
    expected_result = '0.1.2.3.4.255'


class TestFunitHardwareVersion(TestFunitFirmwareVersion):
    FunitClass = HardwareVersion
    call_args = [0, 0x13, 0x15]
