import copy
import json
from unittest import TestCase, skip

import spread_core.bam.entities as e_lib
import spread_core.bam.managers as m_lib
import spread_core.errors.project_errors as core_err
import spread_core.tools.settings_const as C
from spread_core import mqtt
from spread_core.bam.mercury.managers import Mercury
from spread_core.bam.mercury.providers import MercuryElectricMeter
from spread_core.mqtt.spread.variable import Variable
from spread_core.tests import utils
from spread_core.tools.incrementor import Increment

P_ID = 0
M_ID = 1

MERCURY_MANAGER = {
    "type": "Mercury",
    "id": M_ID,
    "attributes": {
        "pollRate": 1000,
        "bus": "Moxa/1"
    }
}

PROVIDER_ATTRIBUTES = {
    "password2": "222222",
    "password1": "111111"
}

CONFIG = {
    "include": [
        "/etc/spread/common.json"
    ],
    "managerID": 1,
    "pollRate": None,
    "useTimestamp": True,
    "communicateTimeout": 1.0,
    "localAdapter": False,
    "levelSyncDelay": 1.0
}


def parse(data):
    pass


class _TestMercury(TestCase):
    @staticmethod
    def start_interface_handler(instance):
        pass

    @classmethod
    def setUpClass(cls) -> None:
        m_lib.config.args = [None]

    def setUp(self) -> None:
        self.manager = Mercury(P_ID, M_ID)
        self.warehouse = self.manager.warehouse

        global parse
        parse = self.manager.parse

        self.manager.set_broker(utils.TestMqttClient())
        parse(MERCURY_MANAGER)
        self.manager.set_interface()

    def append_provider(self, item, fixture=None):
        if fixture:
            fixture(item)

        self.warehouse.add_provider(item)
        self.warehouse.add_binding(item.binding, item)


class TestDaliInitialization(_TestMercury):
    def test_device_count(self):
        self.assertEqual(len(self.warehouse.providers), 0)
        self.assertEqual(len(self.warehouse.bindings), 0)

    def test_name(self):
        self._set_and_parse(C.NAME, "bla bla bla")
        self.assertEqual(self.manager.name, "bla bla bla")

        self._set_and_parse(C.NAME, None, clear=True)
        self.assertEqual(self.manager.name, f'id:{M_ID}')

    def test_bus(self):
        self._set_and_parse(C.BUS, 'bla bla bla', for_attrs=True)
        self.assertEqual(self.manager.bus, 'bla bla bla')

        self._set_and_parse(C.BUS, None, for_attrs=True, clear=True)
        self.assertIsNone(self.manager.bus)

    def test_poll_rate(self):
        self._set_and_parse(C.POLL_RATE, 1234, for_attrs=True)
        self.assertEqual(self.manager.poll_rate, 1.2)
        self.assertEqual(self.manager._def_poll_rate, 1234)

    def test_poll_rate_empty(self):
        self._set_and_parse(C.POLL_RATE, 0, for_attrs=True, clear=True)
        self.assertEqual(self.manager.poll_rate, 0)
        self.assertEqual(self.manager._def_poll_rate, 0)

    def test_empty_data(self):
        try:
            parse(None)
        except TypeError:
            return
        except:
            self.fail("Failed empty data parsing")
        else:
            self.fail("Parsing empty data did not trigger completion!")

    @staticmethod
    def _set_and_parse(key, value, for_attrs=False, clear=False):
        init_data = copy.deepcopy(MERCURY_MANAGER)
        data_for_change = init_data[C.ATTRIBUTES] if for_attrs else init_data
        if clear:
            data_for_change.pop(key, None)
        else:
            data_for_change[key] = value

        parse(init_data)


class TestMercuryConfigurations(_TestMercury):
    def set_and_update(self, key, value, clear=False):
        if clear:
            m_lib.config.pop(key, None)
        else:
            m_lib.config[key] = value

        self.manager.update_config()

    def test_poll_rate(self):
        self.set_and_update(C.POLL_RATE, 3210)
        self.assertEqual(self.manager.poll_rate, 3.2)

        self.set_and_update(C.POLL_RATE, None)
        self.assertEqual(self.manager.poll_rate, 1)

        self.set_and_update(C.POLL_RATE, 0, clear=True)
        self.assertEqual(self.manager.poll_rate, 1)

        self.set_and_update(C.POLL_RATE, None)
        self.assertEqual(self.manager.poll_rate, 1)

    def test_communicate_timeout(self):
        self.set_and_update(C.COMMUNICATE_TIMEOUT, 999)
        self.assertEqual(self.manager.send_interface.timeout, 999)

    def test_session_time(self):
        self.set_and_update(C.SESSION_TIME, 888)
        self.assertEqual(self.manager.send_interface.SESSION_TIME, 888)


class TestMercuryManaging(_TestMercury):
    def setUp(self) -> None:
        super(TestMercuryManaging, self).setUp()
        i = Increment(0)

        for binding_index in range(0, 3):
            self.append_provider(MercuryElectricMeter(i.next(), i.current, PROVIDER_ATTRIBUTES))

    def test_counts(self):
        self.assertEqual(3, len(self.warehouse.providers))

    def test_device(self):
        provider = MercuryElectricMeter(999, 99, PROVIDER_ATTRIBUTES)
        self.add_and_check(provider)
        self.assertIn(provider, self.manager.providers_by_address(99))

    def add_and_check(self, provider):
        self.append_provider(provider)
        self.assertEqual(4, len(self.warehouse.providers))
        self.assertIs(provider, self.warehouse.get_provider(999))


class TestCommandReceiver(_TestMercury):
    @classmethod
    def setUpClass(cls) -> None:
        super(TestCommandReceiver, cls).setUpClass()
        m_lib.Entity.funit_class = utils.TestFunit
        e_lib.Funit = utils.TestFunit

    def setUp(self) -> None:
        super(TestCommandReceiver, self).setUp()
        self.manager.warehouse = utils.TestStore()
        self.reply_topic = mqtt.spread.topic.Reply(P_ID, 'test_session').__str__()
        self.variable = Variable('testValue', session='test_session')

    def test_entity_address(self):
        address = self.manager.get_address('TestFunit')
        self.assertIsInstance(address, mqtt.spread.address.ManagerAddress)
        self.assertEqual(address.type, "Mercury")
        self.assertEqual(address.id, M_ID)
        self.assertEqual(address.funit_type, "TestFunit")

    def test_entity_address_sharp(self):
        address = self.manager.get_address(None)
        self.assertIsInstance(address, mqtt.spread.address.ManagerAddress)
        self.assertEqual(address.type, "Mercury")
        self.assertEqual(address.id, M_ID)
        self.assertEqual(address.funit_type, "#")

    @skip
    def test_invalid_bus(self):
        address = mqtt.spread.address.ProviderAddress(P_ID, "Manager", M_ID, "Provider", 999, "TestFunit")
        self.manager.on_command(mqtt.spread.topic.Set(address), self.variable)

        self.assertEqual(len(self.manager.mqttc.messages), 1)

        message = self.manager.mqttc.messages[0]
        self.assertEqual(message.topic, self.reply_topic)
        self.assertEqual(message.delay, 0)
        self.assertFalse(message.retain)

        data = json.loads(message.data)

        self.assertTrue(data.get('invalid'))
        self.assertEqual(data.get('error'), "Bus is not ready")

    @skip
    def test_manager_command(self):
        pass

    def test_provider_command(self):
        address = mqtt.spread.address.ProviderAddress(P_ID, "Manager", M_ID, "Provider", 999, "TestFunit")

        self.manager.send_interface.ready = True
        with self.assertRaises(utils.Popup) as popup:
            self.manager.on_command(mqtt.spread.topic.Set(address), self.variable)

        self.assertEqual(popup.exception.args[0], 999)


class TestStateReceiver(_TestMercury):
    @staticmethod
    def fixture(provider):
        provider.on_state = utils.Handler.proceed

    def test_state_receiver(self):
        self.append_provider(MercuryElectricMeter(999, 99, PROVIDER_ATTRIBUTES), fixture=self.fixture)
        address = mqtt.spread.address.ProviderAddress(P_ID, "Mercury", M_ID, "MercuryElectricMeter", 999, "SomeState")
        topic = mqtt.spread.topic.State(address)
        try:
            self.manager.on_state(topic, Variable('Success'))
        except utils.Handler as ex:
            self.assertEqual(ex.args[0], "SomeState")
            self.assertEqual(ex.args[1].value, "Success")
        else:
            self.fail(f'Unhandled state')


@skip
class TestExitHandler(_TestMercury):
    pass
