from unittest import TestCase
import asyncio
from spread_core.projector_async_new import Projector
from spread_core.tools import settings
from spread_core import mqtt
from spread_lom.profiles import DimmingProfile
from spread_core.bam.engineries.lighting_area import ButtonPreset
import os
import json
import shutil


class TestProjector(TestCase):
    class Client:
        def __init__(self):
            self.calls = set()

        async def publish(self, topic, payload, retain):
            self.calls.add((topic, payload, retain))

    wrong_cycle = mqtt.project.Cycle(123, 999)
    wrong_trigger = mqtt.project.CycleTrigger(123, 999, 1)
    wrong_provider = mqtt.project.Provider(123, 'Dali', 1, 'DaliDimmer', 999)
    cycle = mqtt.project.Cycle(123, 1)
    cycle_triggers = mqtt.project.Cycle(123, 2)
    trigger = mqtt.project.CycleTrigger(123, 1, 1)
    trigger1 = mqtt.project.CycleTrigger(123, 2, 1)
    trigger2 = mqtt.project.CycleTrigger(123, 2, 2)
    trigger3 = mqtt.project.CycleTrigger(123, 2, 3)
    trigger4 = mqtt.project.CycleTrigger(123, 2, 4)
    button_preset = mqtt.project.Preset(123, ButtonPreset.__name__, 1)
    dimming_profile = mqtt.project.Preset(123, DimmingProfile.__name__, 2)
    provider = mqtt.project.Provider(123, 'Dali', 1, 'DaliDimmer', 2)
    manager = mqtt.project.Hardware(123, 'Dali', 1)

    def setUp(self) -> None:
        settings.config = {
            "localPath": "/var/project/test",
            "projectID": 123,
        }

    def dataset(self):
        self.projector = Projector()
        self.projector.client = self.Client()
        self.projector.verify_mode = 2
        self.projector.ctp = {}
        self.projector.CYCLES = []

        # Clear test dir
        if os.path.exists(settings.config["localPath"]):
            shutil.rmtree(settings.config["localPath"])

        self.cycle_data = {
            'name': 'test1',
            'period': 3,
            'duration': 1,
            'startLevel': 25
        }
        self.cycle_triggers_storage_data = {
            'name': 'test2',
            'period': 7,
            'duration': 1,
            'startLevel': 25,
            "triggers": {
                "1": {"day": 1, "time": "09:00", "level": 50},
                "2": {"day": 2, "time": "09:00", "level": 50},
                "3": {"day": 3, "time": "09:00", "level": 50},
            }
        }
        self.trigger_data = {"day": 0, "time": "09:00", "level": 50}
        self.trigger4_data = {"day": 4, "time": "09:00", "level": 50}
        self.button_preset_data = {
            "handlers": [
                [
                    {
                        "level": 50,
                        "stopTuning": False,
                        "event": "ShortPress",
                        "command": "Level",
                        "setLightingArea": "On"
                    }
                ],
                [
                    {
                        "stopTuning": False,
                        "event": "ShortPress",
                        "command": "Off",
                        "setLightingArea": "Off"
                    }
                ]
            ],
            "name": "Button1",
            "isToggle": True
        }
        self.button_preset_data2 = {
            "handlers": [
                [
                    {
                        "level": 55,
                        "stopTuning": True,
                        "event": "ShortPress",
                        "command": "Level",
                        "setLightingArea": "On"
                    }
                ],
                [
                    {
                        "stopTuning": True,
                        "event": "ShortPress",
                        "command": "Off",
                        "setLightingArea": "Off"
                    }
                ]
            ],
            "name": "Button2",
            "isToggle": True
        }
        self.dimming_profile_data = {
            "items": [
                {
                    "hours": "18",
                    "minutes": "00",
                    "value": 80,
                    "id": 1
                }
            ],
            "name": "time_level",
            "type": "UTC",
            "operand": "Level"
        }
        self.dimming_profile_data2 = {
            "items": [
                {
                    "hours": "03",
                    "minutes": "00",
                    "value": 1,
                    "id": 1
                }
            ],
            "name": "time_scene",
            "type": "UTC",
            "operand": "Scene"
        }
        self.provider_data = {
            "attributes": {
                "binding": "Device",
                "device": 1,
                "group": 1
            },
            "key": "c3ccf369-3766-4733-a443-7acd98a54ecc"
        }
        self.provider_data2 = {
            "attributes": {
                "binding": "Device",
                "device": 1,
                "group": 0
            },
            "key": "458a2f1e-ed37-42a0-a06c-99032e967cfd"
        }

    def test_project_init(self):
        """
            Что проверять:
                ctp             # {entity: entity_data}
                project         # ctp + storage
                SERVERS         # [server_id]
                MANAGERS        # {'manager_id': 'manager_type'}
                PROVIDERS       # {'provider_id': 'provider_type'}
                ENGINERIES      # {'enginery_id': 'enginery_type'}
                LOCATIONS       # [location_id]
                CYCLES          # [cycle_id]
        """
        settings.config = {
            "ctpPath": "spread_core/tests/projector/throne-project_1104.ctp",
            "localPath": "/var/project/test",
            "projectID": 1104,
        }
        self.dataset()
        import spread_core.tests.projector.project_init_template as template
        storage = {
            mqtt.project.Provider(template.id, "Dali", 4770, "DaliDimmer", 178975): self.provider_data,
            mqtt.project.Cycle(template.id, 1): self.cycle_data,
            mqtt.project.CycleTrigger(template.id, 1, 1): self.trigger_data,
            mqtt.project.Preset(template.id, ButtonPreset.__name__, 1): self.button_preset_data,
            mqtt.project.Preset(template.id, DimmingProfile.__name__, 1): self.dimming_profile_data
        }
        project = template.ctp.copy()
        project.update(storage)
        cycle_storage_data = self.cycle_data.copy()
        cycle_storage_data.update({"triggers": {"1": self.trigger_data}})
        self.projector.storage.save(mqtt.project.Provider(template.id, "Dali", 4770, "DaliDimmer", 178975), self.provider_data)
        self.projector.storage.save(mqtt.project.Cycle(template.id, 1), cycle_storage_data)
        self.projector.storage.save(mqtt.project.Preset(template.id, ButtonPreset.__name__, 1), self.button_preset_data)
        self.projector.storage.save(mqtt.project.Preset(template.id, DimmingProfile.__name__, 1), self.dimming_profile_data)

        asyncio.run(self.projector.project_init())
        self.assertEqual(self.projector.ctp, template.ctp)
        self.assertEqual(self.projector.project, project)
        self.assertEqual(self.projector.SERVERS, template.SERVERS)
        self.assertEqual(self.projector.MANAGERS, template.MANAGERS)
        self.assertEqual(self.projector.PROVIDERS, template.PROVIDERS)
        self.assertEqual(self.projector.ENGINERIES, template.ENGINERIES)
        self.assertEqual(self.projector.LOCATIONS, template.LOCATIONS)
        self.assertEqual(self.projector.CYCLES, template.CYCLES)

    def test_data_valid(self):
        self.dataset()
        self.assertTrue(self.projector.data_valid(self.cycle, self.cycle_data))
        if self.projector.verify_mode > 1:
            self.assertFalse(self.projector.data_valid(self.cycle, {'name': 'test'}))

    """
        Positive tests:
            storage_only_entity: (create / change / reset)
                cycle
                trigger
                button_preset
                dimming_profile
            
            ctp entity: (change / reset)
                enginery, manager, provider, location, subginery, user

        Invalid tests:
            create ctp entity (provider)
            create trigger without cycle
            reset not created entity (cycle, trigger, button_preset, dimming_profile, provider)
    """

    def test_create_cycle(self):
        self.dataset()
        asyncio.run(self.projector.set_topic(self.cycle, self.cycle_data.copy()))
        self.assertEqual(self.cycle_data, self.projector.storage.load(self.cycle))
        calls = {
            (str(self.cycle), json.dumps(self.cycle_data), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertIn(self.cycle.entity_id, self.projector.CYCLES)

    def test_create_trigger(self):
        self.dataset()
        self.projector.storage.save(self.cycle, self.cycle_data)
        self.projector.CYCLES = [1]
        asyncio.run(self.projector.set_topic(self.trigger, self.trigger_data))
        self.cycle_data['triggers'] = {}
        self.cycle_data['triggers']['1'] = self.trigger_data
        self.assertEqual(self.cycle_data, self.projector.storage.load(self.cycle))
        calls = {
            (str(self.trigger), json.dumps(self.trigger_data), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertIn(self.cycle.entity_id, self.projector.CYCLES)

    def test_create_trigger_cycle_triggers(self):
        self.dataset()
        self.projector.storage.save(self.cycle_triggers, self.cycle_triggers_storage_data)
        self.projector.CYCLES = [2]
        asyncio.run(self.projector.set_topic(self.trigger4, self.trigger4_data))
        self.cycle_triggers_storage_data['triggers']['4'] = self.trigger4_data
        self.assertEqual(self.cycle_triggers_storage_data, self.projector.storage.load(self.cycle_triggers))
        calls = {
            (str(self.trigger4), json.dumps(self.trigger4_data), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertIn(self.cycle_triggers.entity_id, self.projector.CYCLES)

    def test_change_cycle(self):
        self.dataset()
        self.projector.storage.save(self.cycle_triggers, self.cycle_triggers_storage_data)
        self.projector.CYCLES = [2]
        asyncio.run(self.projector.set_topic(self.cycle_triggers, self.cycle_data.copy()))
        triggers = self.cycle_triggers_storage_data.pop('triggers')
        new_storage_data = self.cycle_data.copy()
        new_storage_data.update({'triggers': triggers})
        self.assertEqual(new_storage_data, self.projector.storage.load(self.cycle_triggers))
        calls = {
            (str(self.cycle_triggers), json.dumps(self.cycle_data), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertIn(self.cycle_triggers.entity_id, self.projector.CYCLES)

    def test_change_trigger(self):
        self.dataset()
        self.projector.storage.save(self.cycle_triggers, self.cycle_triggers_storage_data)
        self.projector.CYCLES = [2]
        asyncio.run(self.projector.set_topic(self.trigger2, self.trigger_data))
        self.cycle_triggers_storage_data['triggers']['2'] = self.trigger_data
        self.assertEqual(self.cycle_triggers_storage_data, self.projector.storage.load(self.cycle_triggers))
        calls = {
            (str(self.trigger2), json.dumps(self.trigger_data), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertIn(self.cycle_triggers.entity_id, self.projector.CYCLES)

    def test_reset_cycle(self):
        self.dataset()
        self.projector.storage.save(self.cycle, self.cycle_data)
        self.projector.CYCLES = [1]
        asyncio.run(self.projector.reset_topic(self.cycle, {}))
        self.assertEqual(None, self.projector.storage.load(self.cycle))
        calls = {
            (str(self.cycle), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertNotIn(self.cycle.entity_id, self.projector.CYCLES)

    def test_reset_cycle_triggers(self):
        self.dataset()
        self.projector.storage.save(self.cycle_triggers, self.cycle_triggers_storage_data)
        self.projector.CYCLES = [2]
        asyncio.run(self.projector.reset_topic(self.cycle_triggers, {}))
        self.assertEqual(None, self.projector.storage.load(self.cycle_triggers))
        calls = {
            (str(self.trigger1), bytes(), True),
            (str(self.trigger2), bytes(), True),
            (str(self.trigger3), bytes(), True),
            (str(self.cycle_triggers), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertNotIn(self.cycle_triggers.entity_id, self.projector.CYCLES)

    def test_reset_trigger(self):
        self.dataset()
        self.projector.storage.save(self.cycle_triggers, self.cycle_triggers_storage_data)
        self.projector.CYCLES = [2]
        asyncio.run(self.projector.reset_topic(self.trigger2, {}))
        self.cycle_triggers_storage_data['triggers'].pop('2')
        self.assertEqual(self.cycle_triggers_storage_data, self.projector.storage.load(self.cycle_triggers))
        calls = {
            (str(self.trigger2), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertIn(self.cycle_triggers.entity_id, self.projector.CYCLES)

    def test_create_button_preset(self):
        self.dataset()
        asyncio.run(self.projector.set_topic(self.button_preset, self.button_preset_data.copy()))
        self.assertEqual(self.button_preset_data, self.projector.storage.load(self.button_preset))
        calls = {
            (str(self.button_preset), json.dumps(self.button_preset_data), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)

    def test_change_provider(self):
        self.dataset()
        self.projector.ctp[self.provider] = self.provider_data
        self.projector.MANAGERS = {self.manager.entity_id: self.manager.entity_type}
        asyncio.run(self.projector.set_topic(self.provider, self.provider_data2))
        self.assertEqual(self.provider_data2, self.projector.storage.load(self.provider))
        calls = {
            (str(self.provider), json.dumps(self.provider_data2), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)

    def test_change_provider_ctp(self):
        self.dataset()
        self.projector.ctp[self.provider] = self.provider_data
        self.projector.MANAGERS = {self.manager.entity_id: self.manager.entity_type}
        self.projector.storage.save(self.provider, self.provider_data2)
        asyncio.run(self.projector.set_topic(self.provider, self.provider_data))
        self.assertEqual(None, self.projector.storage.load(self.provider))
        calls = {
            (str(self.provider), json.dumps(self.provider_data), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)

    def test_reset_provider(self):
        self.dataset()
        self.projector.ctp[self.provider] = self.provider_data
        self.projector.MANAGERS = {self.manager.entity_id: self.manager.entity_type}
        self.projector.storage.save(self.provider, self.provider_data2)
        asyncio.run(self.projector.reset_topic(self.provider, {}))
        calls = {
            (str(self.provider), json.dumps(self.provider_data), True)
        }
        self.assertEqual(None, self.projector.storage.load(self.provider))
        self.assertSetEqual(calls, self.projector.client.calls)

    def test_set_wrong_trigger(self):
        self.dataset()
        with self.assertRaises(Exception):
            asyncio.run(self.projector.set_topic(self.wrong_trigger, self.trigger_data))
        self.assertEqual(None, self.projector.storage.load(mqtt.project.Cycle(123, self.wrong_trigger.cycle_id)))
        self.assertSetEqual(set(), self.projector.client.calls)
        self.assertNotIn(self.wrong_trigger.cycle_id, self.projector.CYCLES)

    def test_set_wrong_provider(self):
        self.dataset()
        self.projector.ctp[self.provider] = self.provider_data
        self.projector.MANAGERS = {self.manager.entity_id: self.manager.entity_type}
        with self.assertRaises(Exception):
            asyncio.run(self.projector.set_topic(self.wrong_provider, self.provider_data))
        self.assertEqual(None, self.projector.storage.load(self.wrong_provider))
        self.assertSetEqual(set(), self.projector.client.calls)

    def test_wrong_reset(self):
        self.dataset()
        self.projector.storage.save(self.cycle, self.cycle_data)
        self.projector.CYCLES = [1]
        with self.assertRaises(Exception):
            asyncio.run(self.projector.reset_topic(self.cycle, 'wrong_reset_data'))
        self.assertEqual(self.cycle_data, self.projector.storage.load(self.cycle))
        self.assertSetEqual(set(), self.projector.client.calls)
        self.assertIn(self.cycle.entity_id, self.projector.CYCLES)

    def test_reset_wrong_cycle(self):
        self.dataset()
        asyncio.run(self.projector.reset_topic(self.wrong_cycle, {}))
        calls = {
            (str(self.wrong_cycle), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)

    def test_reset_wrong_trigger(self):
        self.dataset()
        asyncio.run(self.projector.reset_topic(self.wrong_trigger, {}))
        calls = {
            (str(self.wrong_trigger), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)

    def test_reset_trigger_out_of_cycle(self):
        self.dataset()
        self.projector.storage.save(self.cycle, self.cycle_data)
        self.projector.CYCLES = [1]
        asyncio.run(self.projector.reset_topic(self.trigger, {}))
        self.assertEqual(self.cycle_data, self.projector.storage.load(self.cycle))
        calls = {
            (str(self.trigger), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertIn(self.cycle.entity_id, self.projector.CYCLES)

    def test_reset_trigger_out_of_cycle_triggers(self):
        self.dataset()
        self.projector.storage.save(self.cycle_triggers, self.cycle_triggers_storage_data)
        self.projector.CYCLES = [2]
        asyncio.run(self.projector.reset_topic(self.trigger4, {}))
        self.assertEqual(self.cycle_triggers_storage_data, self.projector.storage.load(self.cycle_triggers))
        calls = {
            (str(self.trigger4), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)
        self.assertIn(self.cycle_triggers.entity_id, self.projector.CYCLES)

    def test_reset_wrong_button_preset(self):
        self.dataset()
        asyncio.run(self.projector.reset_topic(self.button_preset, {}))
        calls = {
            (str(self.button_preset), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)

    def test_reset_wrong_dimming_profile(self):
        self.dataset()
        asyncio.run(self.projector.reset_topic(self.dimming_profile, {}))
        calls = {
            (str(self.dimming_profile), bytes(), True)
        }
        self.assertSetEqual(calls, self.projector.client.calls)

    def test_reset_wrong_provider(self):
        self.dataset()
        self.projector.ctp[self.provider] = self.provider_data
        self.projector.MANAGERS = {self.manager.entity_id: self.manager.entity_type}
        with self.assertRaises(Exception):
            asyncio.run(self.projector.set_topic(self.wrong_provider, self.provider_data))
        self.assertEqual(None, self.projector.storage.load(self.wrong_provider))
        self.assertSetEqual(set(), self.projector.client.calls)
