from ..address import Address, Short, Group, Instance, InstanceType, InstanceNumber, InstanceGroup


class Event:
    @classmethod
    def of_frame(cls, frame):
        if not isinstance(frame, (bytes, bytearray)):
            raise ValueError('Frame must be bytes')
        elif len(frame) != 3:
            raise ValueError('Frame size is {}. Expected frame size is 3'.format(len(frame)))
        elif frame[0] & 1:
            raise ValueError('16th bit of frame is set. Frame is not Even!')

        source1_value = (frame[0] >> 1) & 31
        source2_value = (frame[1] >> 2) & 31
        value = int.from_bytes(frame, 'big') & 1023

        if frame[0] >> 7 == 0:
            short_address = Short((frame[0] >> 1) & 63)

            if frame[1] >> 7 == 0:
                return cls(short_address, InstanceType(source2_value), value)

            elif frame[1] >> 7 == 1:
                return cls(short_address, InstanceNumber(source2_value), value)

        elif frame[0] >> 6 == 0b10:

            if frame[1] >> 7 == 0:
                return cls(Group(source1_value), InstanceType(source2_value), value)

            elif frame[1] >> 7 == 1:
                return cls(InstanceType(source1_value), InstanceNumber(source2_value), value)

        elif frame[0] >> 6 == 0b11:

            if frame[1] >> 7 == 0:
                return cls(InstanceGroup(source1_value), InstanceType(source2_value), value)

        raise ValueError("Unknown event scheme")

    def __init__(self, source1, source2, value):
        if isinstance(source1, InstanceType):
            self._owner_type = source1.value
        elif isinstance(source2, InstanceType):
            self._owner_type = source2.value
        else:
            self._owner_type = None

        self.source1 = source1
        self.source2 = source2
        self.value = value
        self._data = None

    @property
    def data(self):
        if self._data is None:
            if self._owner_type == 1:
                self._data = ButtonEvent(self.value)
            elif self._owner_type == 3:
                self._data = PresenceEvent(self.value)
            elif self._owner_type == 4:
                self._data = LuminosityEvent(self.value)
            else:
                self._data = EventData(self.value)

        return self._data

    @property
    def owner_type(self):
        return self._owner_type

    @owner_type.setter
    def owner_type(self, owner_type):
        self._data = None
        self._owner_type = owner_type

    def __str__(self):
        return f'{self.data.__class__.__name__}({self.source1}, {self.source2}): {self.data.value}'

    def __repr__(self):
        return self.__str__()


class EventData:
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    @staticmethod
    def raw_of_value(value):
        if value < 0:
            raise ValueError("Value must be positive int")
        if value > 1023:
            raise ValueError("Value must be lower than 1024")

        return value


class LuminosityEvent(EventData):
    pass


class PresenceEvent(EventData):
    """
    00 0000 ---0 - No movement detected. Corresponding trigger is the ‘No movement’ trigger.
    00 0000 ---1 - Movement detected. Corresponding trigger is the ‘Movement’ trigger.
    00 0000 -00- - Vacant. The area has become vacant. Corresponding trigger is the ‘Vacant’ trigger.
    00 0000 -10- - Still vacant. The area is still vacant. The event occurs at regular intervals as long as
                        the vacant condition holds. Corresponding trigger is the ‘Repeat’ trigger.
    00 0000 -01- - Occupied The area has become occupied. Corresponding trigger is the ‘Occupied’ trigger.
    00 0000 -11- - Still occupied. The area is still occupied. The event occurs at regular intervals as long as
                        the occupied condition holds. Corresponding trigger is the ‘Repeat’ trigger.
    00 0000 0--- - Presence sensor. The current event is triggered by a presence based sensor.
    00 0000 1--- - Movement sensor. The current event is triggered by a movement based sensor.
    """

    @property
    def has_move(self): return self._value & 0x1 == 0x1

    @property
    def is_occupied(self): return self._value & 0x2 == 0x2

    @property
    def is_still_occupied(self): return self._value & 0x4 == 0x4 and self.is_occupied

    @property
    def is_still_vacant(self): return self._value & 0x4 == 0x4 and not self.is_occupied

    @property
    def by_presence(self): return self._value & 0x8 == 0x0

    @property
    def by_movement(self): return self._value & 0x8 == 0x8

    def __str__(self):
        return '[{}]: {} ({})'.format(self.__class__.__name__, self.value, bin(self._value)[2:].rjust(5, '0'))

    @property
    def value(self): return self.has_move or self.is_occupied

    @staticmethod
    def raw_of_value(value: bool):
        if value:
            return 0b11

        return 0


class ButtonEvent(EventData):
    """
    buttonReleased	00 0000 0000		The button is released.
    buttonPressed 	00 0000 0001		The button is pressed.
    shortPress		00 0000 0010        The button is pressed and released, without being pressed quickly again (in
                                        case double press is enabled), or the button is pressed and quickly released
                                        (in case double press is disabled).
    doublePress		00 0000 0101		The button is pressed and released, quickly followed by another buttonpress.
    longPressStart	00 0000 1001		The button is pressed without releasing it.
    longPressRepeat	00 0000 1011		Following a long press start condition, the button is still pressed. The event
                                        occurs at regular intervals as long as the condition holds.
    longPressStop	00 0000 1100		Following a long press start condition, the button is released.
    buttonFree		00 0000 1110		The button has been stuck and is now released.
    buttonStuck		00 0000 1111		The button has been pressed for a very long time and is assumed stuck.
    """

    BUTTON_RELEASED     = 'ButtonReleased'
    BUTTON_PRESSED      = 'ButtonPressed'
    SHORT_PRESS         = 'ShortPress'
    DOUBLE_PRESS        = 'DoublePress'
    LONG_PRESS_START    = 'LongPressStart'
    LONG_PRESS_REPEAT   = 'LongPressRepeat'
    LONG_PRESS_STOP     = 'LongPressStop'
    BUTTON_FREE         = 'ButtonFree'
    BUTTON_STUCK        = 'ButtonStuck'
    DEFAULT             = 'UnknownButtonEvent'

    STATES = {
        0: BUTTON_RELEASED,
        1: BUTTON_PRESSED,
        2: SHORT_PRESS,
        5: DOUBLE_PRESS,
        9: LONG_PRESS_START,
        11: LONG_PRESS_REPEAT,
        12: LONG_PRESS_STOP,
        14: BUTTON_FREE,
        15: BUTTON_STUCK,
    }

    @property
    def value(self):
        return self.STATES.get(self._value, self.DEFAULT)

    @staticmethod
    def raw_of_value(value: str):
        for raw_value, value in ButtonEvent.STATES.items():
            if value == value:
                return raw_value

        raise ValueError("%s is unexpected event value" % value)


def frame_of_sources(source1: [Short, Group, InstanceType, InstanceGroup],
                     source2: [InstanceType, InstanceNumber],
                     value: int) -> bytes:
    """
    Generates an event-frame of given sources with given value

    :param source1: first part of the EventScheme
    :param source2: second part of the EventScheme
    :param value: the 10 bits long value
    :return: frame of bytes
    """

    if not isinstance(value, int) or value not in range(1024):
        raise ValueError("The event value is one of [0..1023]")

    if not is_available_schema(source1, source2):
        raise ValueError("Unsuitable event scheme")

    _source0_flag = not isinstance(source1, Short)
    _source1_flag = isinstance(source1, InstanceGroup)
    _source2_flag = isinstance(source2, InstanceNumber)

    if isinstance(source1, Short):
        _source1 = source1.address
    elif isinstance(source1, Group):
        _source1 = source1.group
    else:
        assert isinstance(source1, Instance)
        _source1 = source1.value

    _source2 = source2.value

    frame = sum((
        _source0_flag << 23,
        _source1_flag << 22,
        _source2_flag << 15,
        _source1 << 17,
        _source2 << 10,
        value
    ))

    return frame.to_bytes(3, 'big')


def is_available_schema(source1: Address, source2: Instance):
    """Check if EventSchema is available"""

    if isinstance(source2, InstanceType):
        return isinstance(source1, (Short, Group, InstanceGroup))

    if isinstance(source2, InstanceNumber):
        return isinstance(source1, (Short, InstanceType))

    return False


def event_of_sources(source1, source2, value):
    return Event.of_frame(frame_of_sources(source1, source2, value))
