import random
import string
from functools import reduce
from typing import Generator

from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic import TemplateView
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _

import spread_core.mqtt as mqtt
from spread_core.bam import engineries
from spread_dashboard.project import bam
from spread_dashboard.responses import Error
from spread_dashboard.commissioning import models
from spread_dashboard.scheduler.models import LightFactor


class CommissioningIndex(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/index.html'

    def get(self, request, *args, **kwargs):
        return self.render_to_response({
            'locations': bam.models.Location.as_tree(),
            'projectName': bam.models.Header.objects.get().name,
            'broker': bam.broker_auth()
        })


class Intro(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/intro.html'

    def get(self, request, *args, **kwargs):
        return self.render_to_response({'navigations': ({'name': _('intro')},)})


class Glossary(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/glossary.html'

    def get(self, request, *args, **kwargs):
        return self.render_to_response({'navigations': ({'name': _('glossary')},)})


class CommissioningLocation(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/location.html'

    def get(self, request, *args, location_id=None, **kwargs):
        context = make_location_context(location_id)
        context.update(
            {
                'title_1_href': f'commissioning/admin/{location_id}/',
                'projectName': bam.models.Header.objects.get().name,
                'broker': bam.broker_auth(),
                'dateFrom': request.GET.get('dateFrom', ''),
                'dateTo': request.GET.get('dateTo', ''),
                'show_calendar': True,
            }
        )
        return self.render_to_response(context)


class CommissioningArea(LoginRequiredMixin, TemplateView):
    template_name = 'commissioning/area.html'

    def get(self, request, *args, location_id: int = None, area_id: int = None, **kwargs):
        context = make_area_context(location_id, area_id)
        context.update({
            'title_1_href': f'commissioning/admin/{location_id}/{area_id}',
            'projectName': bam.models.Header.objects.get().name,
            'broker': bam.broker_auth(),
            'dateFrom': request.GET.get('dateFrom', ''),
            'dateTo': request.GET.get('dateTo', ''),
            'show_calendar': True,
        })
        return self.render_to_response(context)


def enginery_count_on_location(location: bam.models.Location, engineries_types: (list, tuple)):
    return len(list(location.engineries(engineries_types)))


def enginery_count_by_subginery(location: bam.models.Location, engineries_types: (list, tuple)):
    """ Метод, который считает светильники в локации по Subginery"""
    objects = bam.models.Subginery.objects.filter(locationID=location.id, type__in=('Lighting',))
    objects = objects and objects[0].engineries
    if objects:
        objects = bam.models.Enginery.objects.filter(id__in=objects, type__in=engineries_types)

    return len(objects)


def enginery_count_on_location_with_sublocations(location: bam.models.Location,
                                                 engineries_types: (list, tuple),
                                                 count_root: bool=True):
    count = len(list(location.engineries(engineries_types))) if count_root else 0

    for sub_location in location.sub_locations:
        count += enginery_count_on_location_with_sublocations(sub_location, engineries_types)

    return count


def engineries_on_area(area: bam.models.Enginery, ingredient_types: str) -> Generator[None, dict, None]:
    """
        Возвращает генератор инженерных объектов световой зоны.

        :param area: модель световой зоны
        :param ingredient_types: типы ингредиентов, которые должны попасть в выборку
        :return: Generator[None, dict, None]
    """
    for ingr in ingredients_of_area(area.id, ingredient_types=[ingredient_types]):
        enginery = ingr.enginery
        provider = ingr.provider
        yield {
            'id': enginery.id,
            'state': provider.state(''),
            'chart': mqtt.spread.topic.Renew(
                mqtt.spread.address.EngineryAddress(
                    bam.models.Header.objects.get().project_id,
                    enginery.type,
                    enginery.id,
                    'Chart'
                )
            ),
            'power': enginery.attributes.get('power') if isinstance(enginery.attributes, dict) else None
        }


def buttons_on_area(area: bam.models.Enginery) -> Generator[None, dict, None]:
    """
        Возвращает генератор выключателей световой зоны.

        :param area: модель световой зоны
        :return: Generator[None, dict, None]
    """
    ingredients = bam.models.RecipeIngredient.objects.filter(enginery=area.id, ingredient_type='Button')
    for ingr in ingredients:
        yield {
            'provider_id': ingr.provider_id,
            'state': ingr.provider.state('')
        }


def location_chart_topic(location):
    try:
        # id = bam.models.Subginery.objects.get(type='Lighting', locationID=location.id).id
        return mqtt.spread.topic.Renew(
            mqtt.spread.address.SubgineryAddress(
                bam.models.Header.objects.get().project_id,
                location.id,
                'Lighting',
                'Chart'
            )
        )
    except bam.models.Subginery.DoesNotExist:
        return ''


def ingredients_of_area(area_id, ingredient_types='all'):
    if ingredient_types == 'all':
        ingredient_types = set(bam.models.RecipeIngredient.objects
                               .filter(~Q(ingredient_type='Solid')).values('ingredient_type'))
    if not isinstance(ingredient_types, (list, set, tuple)):
        raise Error('Параметр ingredient_types должен быть типа (list, set, tuple)')
    providers = bam.models.RecipeIngredient.objects.filter(enginery=area_id,
                                                           ingredient_type__in=ingredient_types).values('provider_id')
    return bam.models.RecipeIngredient.objects.filter(Q(provider_id__in=providers) &
                                                      ~Q(ingredient_type__in=ingredient_types))


def sub_areas(loc: bam.models.Location) -> set:
    areas = set(loc.engineries((engineries.LightingArea.__name__,)))
    for sub_loc in loc.sub_locations:
        areas |= sub_areas(sub_loc)
    return areas


def power_on_area(area: bam.models.Enginery):
    """ Суммарная максимальная мощность, потребляемая в световой зоне """
    consumers = ingredients_of_area(area, ingredient_types=('Actuation',))
    res = reduce(lambda power, item: power + item.enginery.attributes.get('power', 0)
        if isinstance(item.enginery.attributes, dict) else 0, consumers, 0)
    return res


def power_on_location(loc: bam.models.Location):
    """ Суммарная максимальная мощность, потребляемая на локации """
    consumers = engineries_on_location(loc, engineries.light_types)
    return reduce(lambda power, item: power + item.attributes.get('power', 0)
        if isinstance(item.attributes, dict) else 0, consumers, 0)


def engineries_on_location(loc: bam.models.Location, types):
    if not isinstance(types, (list, set, tuple)):
        raise Error('Параметр types должен быть типа (list, set, tuple)')
    engineries = bam.models.Subginery.objects.filter(locationID=loc.id).values('engineries')
    engineries = reduce(lambda lst, item: lst + item.get('engineries'), engineries, [])  # в плоский список
    return bam.models.Enginery.objects.filter(id__in=engineries, type__in=types)


def make_location_context(location_id: int, counter: int = 0, level: int = 0) -> dict:
    project_id = bam.models.Header.objects.get().project_id
    media = f'{settings.MEDIA_URL}{project_id}/media/commissioning'
    try:
        loc = bam.models.Location.objects.get(id=location_id)

        navigations = (
            {
                'id': parent_id,
                'name': bam.models.Location.objects.get(id=parent_id).name
            }
            for parent_id in loc.navigations
        )

        areas = list(loc.engineries((engineries.LightingArea.__name__,)))

        strategy = {}
        for row in models.CommissioningStrategy.objects.filter(area__in=sub_areas(loc)):
            strategy.update(row.strategies(applied=True))
        # определение суммы светильников всех подлокаций без учета светильниов текущего уровня ¯\_(ツ)_/¯
        total_actuators_count = enginery_count_on_location_with_sublocations(loc, engineries.light_types) \
                                - enginery_count_on_location(loc, engineries.light_types)
        total_presences_count = enginery_count_on_location_with_sublocations(loc, (engineries.PresenceSensor.__name__,)) \
                                - enginery_count_on_location(loc, (engineries.PresenceSensor.__name__,))
        total_luminosities_count = enginery_count_on_location_with_sublocations(loc, (engineries.LightSensor.__name__,)) \
                                - enginery_count_on_location(loc, (engineries.LightSensor.__name__,))

        location = {
            'id': loc.id,
            'name': loc.name,
            'counter': counter,
            'margin': level * 2,
            'parent': loc.navigations[-2] if len(loc.navigations) > 1 else None,
            'preview': f'{media}/{location_id}.jpg?v={random_str()}',
            'chart': location_chart_topic(loc),
            'locations': [
                {
                    'id': sub_location.id,
                    'name': sub_location.name,
                    'actuators_count': enginery_count_on_location_with_sublocations(sub_location, engineries.light_types),
                    'presences_count': enginery_count_on_location_with_sublocations(sub_location, (engineries.PresenceSensor.__name__,)),
                    'luminosities_count': enginery_count_on_location_with_sublocations(sub_location, (engineries.LightSensor.__name__,)),
                    'chart': location_chart_topic(sub_location)
                }
                for sub_location in loc.sub_locations
            ],
            'areas': [
                {
                    'id': area.id,
                    'name': area.v_name,
                    'actuators': list(engineries_on_area(area, 'Actuation'))
                } for area in areas
            ],
            'strategy': strategy,
            'actuators_count': enginery_count_by_subginery(loc, engineries.light_types),
            'presences_count': enginery_count_by_subginery(loc, (engineries.PresenceSensor.__name__,)),
            'luminosities_count': enginery_count_by_subginery(loc, (engineries.LightSensor.__name__,)),
            'total_actuators_count': total_actuators_count,
            'total_presences_count': total_presences_count,
            'total_luminosities_count': total_luminosities_count,
            'total_power': power_on_location(loc)
        }
        return {
            'location': location,
            'navigations': navigations
        }
    except bam.models.Location.DoesNotExist:
        return Error(f'location {location_id} does not exist')


def make_area_context(location_id: int, area_id: int, counter: int = 0, level: int = 0) -> dict:
    project_id = bam.models.Header.objects.get().project_id
    media = f'{settings.MEDIA_URL}{project_id}/media/commissioning'
    try:
        loc = bam.models.Location.objects.get(id=location_id)
        area = bam.models.Enginery.objects.get(id=area_id)

        navigations = (
            *(
                {
                    'id': parent_id,
                    'name': bam.models.Location.objects.get(id=parent_id).name
                }
                for parent_id in loc.navigations
            ),
            {
                'id': area_id,
                'name': area.v_name
            }
        )

        try:
            strategy = models.CommissioningStrategy.objects.get(area=area).strategies(applied=True)
        except models.CommissioningStrategy.DoesNotExist:
            strategy = ''

        # по идее, на 1 зону - 1 датчик с 1 коэффициентом
        light_factor = LightFactor.objects.filter(area=area)
        # если коэф-т не задан, помножим все на 0, чтоб обратить на это внимание
        light_factor = light_factor.values('value')[0]['value'] if light_factor else 0

        # рандомная строка, чтоб картинки не кешировались в браузере
        rand = random_str()

        return {
            'navigations': navigations,
            'margin': level * 2,
            'counter': counter,
            'name': area.v_name,
            'area': {
                'id': area_id,
                'name': area.v_name,
                'parent': location_id,
                'preview_lights': f'{media}/{location_id}_{area_id}_1.jpg?v={rand}',
                'preview_sensors': f'{media}/{location_id}_{area_id}_2.jpg?v={rand}',
                'actuators': list(engineries_on_area(area, 'Actuation')),
                'presences': list(engineries_on_area(area, 'Presence')),
                'luminosities': list(engineries_on_area(area, 'Luminosity')),
                'buttons': list(buttons_on_area(area)),
                'state': area.state(''),
                'total_power': power_on_area(area),
                'strategy': strategy,
                'light_factor': light_factor,
                'margin': level * 2,
                'counter': counter,
            },
        }

    except bam.models.Subginery.DoesNotExist:
        return Error('Subginery does not exist')
    except bam.models.Location.DoesNotExist:
        return Error('location <%s> does not exist' % location_id)


def random_str() -> str:
    symbols = string.ascii_letters + string.digits
    return "".join(random.choices(symbols, k=10))
