import os.path
import os

from django.contrib.auth.decorators import login_required
from django.shortcuts import render, redirect
from django.conf import settings
from django import forms

from spread_dashboard.scheduler.models import CommissioningStrategy, LightFactor
from spread_dashboard.project import models, bam
from spread_dashboard.responses import Error
from .models import StrategyForm
from .views import ingredients_of_area, random_str


def index(request):
    return render(request, 'commissioning/editor/index.html', context={
        'title_1_href': f'commissioning/',
        'locations': models.Location.as_tree(),
        'projectName': models.Header.objects.get().name,
    })


@login_required
def of_location(request, location_id: int):
    project_id = bam.models.Header.objects.get().project_id
    media_dir = f'{settings.MEDIA_ROOT}{project_id}/media/commissioning'
    media = f'{settings.MEDIA_URL}{project_id}/media/commissioning'
    if request.method == 'POST':
        for key, file in request.FILES.items():
            save_preview(f'{media_dir}/{location_id}.jpg', file)

        return redirect(request.path)

    try:
        loc = models.Location.objects.get(id=location_id)
        navigations = (
            {
                'id': parent_id,
                'name': models.Location.objects.get(id=parent_id).name
            }
            for parent_id in loc.navigations
        )

        return render(request, 'commissioning/editor/location.html', context={
            'title_1_href': f'commissioning/{location_id}/',
            'navigations': navigations,
            'location': {
                'name': loc.name,
                'preview': f'{media}/{location_id}.jpg?v={random_str()}',
            }
        })

    except models.Location.DoesNotExist:
        return Error('location <%s> does not exist' % location_id)


@login_required
def of_area(request, location_id: int, area_id: int):
    project_id = bam.models.Header.objects.get().project_id
    media_dir = f'{settings.MEDIA_ROOT}{project_id}/media/commissioning'
    media = f'{settings.MEDIA_URL}{project_id}/media/commissioning'
    if request.method == 'POST':
        if 'file' in request.FILES:
            for key, file in request.FILES.items():
                save_preview(f'{media_dir}/{location_id}_{area_id}_{request.POST.get("index")}.jpg', file)

        if 'strategy' in request.POST:
            form = StrategyForm(request.POST, instance=CommissioningStrategy.objects.get(area_id=area_id))
            if form.is_valid():
                if form.changed_data:
                    form.save()

        if 'light_factors' in request.POST:
            formset = get_light_factor_formset(request.POST, area_id=area_id)
            if formset.is_valid():
                formset.save()

        return redirect(request.path)

    try:
        loc = models.Location.objects.get(id=location_id)
        area = models.Enginery.objects.get(id=area_id)

        navigations = (
            *(
                {
                    'id': parent_id,
                    'name': models.Location.objects.get(id=parent_id).name
                }
                for parent_id in loc.navigations
            ),
            {
                'id': area_id,
                'name': area.v_name
            }
        )

        strategy = StrategyForm(instance=CommissioningStrategy.objects.get_or_create(area_id=area_id)[0])
        light_factors = get_light_factor_formset(area_id=area_id)
        # рандомная строка, чтоб картинки не кешировались в браузере
        rand = random_str()

        return render(request, 'commissioning/editor/area.html', context={
            'title_1_href': f'commissioning/{location_id}/{area_id}',
            'navigations': navigations,
            'area': {
                'name': area.v_name,
                'preview_lights': f'{media}/{location_id}_{area_id}_1.jpg?v={rand}',
                'preview_sensors': f'{media}/{location_id}_{area_id}_2.jpg?v={rand}',
                'strategy': strategy,
                'light_factors': light_factors,
            },
            'broker': bam.broker_auth()
        })

    except models.Subginery.DoesNotExist:
        return Error('Subginery does not exist')

    except models.Location.DoesNotExist:
        return Error('location <%s> does not exist' % location_id)


def save_preview(name, file):
    mkdir_or_chmod(name)
    with open(name, 'wb') as dest:
        if file.multiple_chunks:
            for c in file.chunks():
                dest.write(c)
        else:
            dest.write(file.read())

        dest.close()


def mkdir_or_chmod(path: str):
    dirname = os.path.dirname(path)
    if not os.path.exists(dirname):
        os.makedirs(dirname, mode=0o700, exist_ok=True)
    elif not os.access(dirname, os.R_OK | os.W_OK | os.X_OK):
        os.chmod(dirname, 0o700)


def get_light_factor_formset(*args, area_id: int=None):
    """ Создает набор форм для редактирования коэффициента отношения освещенности к показаниям датчика
        :param area_id: int - идентификатор световой зоны (PK)
        :return: LightFactorFormSet """
    if not area_id:
        raise Error('Не передан параметр area_id')

    sensors = []
    extra = 0  # по умолчанию не отображаем формы для добавления новых записей
    queryset = LightFactor.objects.filter(area_id=area_id)
    initial_values = []
    if not args:
        sensors = ingredients_of_area(area_id, ['Luminosity'])
    # если не все датчики в зоне заведены, то создаем формы для них
    if len(queryset) < len(sensors):
        initial_values = [{'area': area_id,
                           'enginery': sensor.enginery,
                           'current_luminosity': sensor.enginery.state('CurrentLuminosity')}
                          for sensor in sensors if sensor.enginery_id not in queryset.values('enginery_id')]
        # вычисляем число форм для новых датчиков
        extra = len(sensors) - len(queryset)

    LightFactorFormSet = forms.modelformset_factory(LightFactor,
                                                    fields=('value', 'area', 'enginery'),
                                                    extra=extra,
                                                    widgets={
                                                        'value': forms.NumberInput(
                                                            attrs={'readonly': True, 'style': 'text-align: right'}
                                                        ),
                                                        'area': forms.NumberInput(attrs={'hidden': True}),
                                                        'enginery': forms.NumberInput(attrs={'hidden': True}),
                                                    }
                                                    )

    return LightFactorFormSet(*args, queryset=queryset, initial=initial_values)
