from django.db import models
from jsonfield import JSONField
from spread_core import mqtt


class Entity(models.Model):
    class Meta:
        abstract = True

    id = models.PositiveIntegerField(primary_key=True)
    type = models.CharField(max_length=255)
    name = models.CharField(max_length=255, blank=True)
    attributes = JSONField()
    key = models.CharField(max_length=64)

    @property
    def v_name(self): return self.name if self.name else 'id:%s' % self.id

    @property
    def name_not_star(self):
        if self.name.endswith('*'):
            return self.name[:-1]
        return self.v_name

    def serialize(self) -> dict:
        return {field.attname: getattr(self, field.attname) for field in self._meta.fields if not field.auto_created}

    @staticmethod
    def separator(entity: dict) -> dict: return entity

    def mqtt_address(self, funit):
        raise NotImplementedError()

    def state(self, funit):
        return mqtt.spread.topic.State(
            self.mqtt_address(funit)
        )

    def renew(self, funit):
        return mqtt.spread.topic.Renew(
            self.mqtt_address(funit)
        )

    def set(self, funit):
        return mqtt.spread.topic.Set(
            self.mqtt_address(funit)
        )


class Header(models.Model):
    version = models.CharField(max_length=4)
    firmware = JSONField()
    project = JSONField()

    @property
    def project_id(self) -> int:
        return self.project['id']

    @property
    def root_location(self) -> int:
        return self.project['rootLocationID']

    @property
    def title(self) -> int:
        return self.project['title']

    @property
    def name(self) -> int:
        return self.project['name']


class Manager(Entity):
    serverID = models.PositiveIntegerField()


class Enginery(Entity):
    recipe = JSONField(blank=True)
    label = models.CharField(max_length=16)

    @staticmethod
    def get_scheduled(loc_id):
        return Enginery.objects.filter(
            name__endswith='*',
            id__in=Subginery.objects.get(
                locationID=loc_id,
                type='Lighting').engineries
        )

    def mqtt_address(self, funit):
        return mqtt.spread.address.EngineryAddress(
            Header.objects.get().project_id,
            self.type,
            self.id,
            funit
        )


class Provider(Entity):
    manager = models.ForeignKey(Manager, on_delete=models.deletion.CASCADE, db_column='managerID')

    def mqtt_address(self, funit):
        return mqtt.spread.address.ProviderAddress(
            Header.objects.get().project_id,
            self.manager.type,
            self.manager.id,
            self.type,
            self.id,
            funit
        )


class Location(Entity):
    scheduled = models.BooleanField()
    navigations = JSONField(default=[], blank=True)
    models = JSONField(default=[], blank=True)
    type = None

    @staticmethod
    def separator(entity: dict) -> dict:
        scheduled = False
        if entity.get('name', '').endswith('*'):
            entity['name'] = entity.get('name', '')[:-1].strip()
            scheduled = True

        return {**{k: entity[k] for k in entity if k in ['id', 'name', 'key', 'navigations', 'models']}, 'scheduled': scheduled}

    def engineries(self, available_types):
        engineries = []
        for model in self.models:
            for ctrl in model.get('controls', []):
                if ctrl.get('clickable', False) and str(ctrl.get('engineryID', None)).isdigit():
                    engineries.append(int(ctrl['engineryID']))

        yield from Enginery.objects.filter(id__in=engineries, type__in=available_types)

    @property
    def sub_locations(self):
        for sub_location in Location.objects.filter(navigations__contains=self.id):
            if sub_location.navigations[:-1] == self.navigations:
                yield sub_location

    def __str__(self): return '%s(%s)' % (self.name, self.id)

    @staticmethod
    def as_tree() -> list:
        def make_item(location: Location) -> dict:
            return {
                'id': location.id,
                'name': location.name,
                'locations': [make_item(sub) for sub in location.sub_locations]
            }

        header = Header.objects.get()
        result = [make_item(Location.objects.get(id=header.root_location))]

        return result


class Subginery(Entity):
    locationID = models.PositiveIntegerField()
    engineries = JSONField(blank=True)


class RecipeIngredient(models.Model):
    id = models.AutoField(primary_key=True)
    enginery = models.ForeignKey(Enginery, on_delete=models.deletion.CASCADE)
    provider = models.ForeignKey(Provider, on_delete=models.deletion.CASCADE)
    index = models.PositiveIntegerField()
    ingredient_type = models.CharField(max_length=255, blank=True)
    recipe_type = models.CharField(max_length=255, blank=True)

    @staticmethod
    def fill_recipes():
        """ Заполняет таблицу при обновлении проекта """
        engineries = Enginery.objects.exclude(recipe="")
        for enginery in engineries:
            recipe = enginery.recipe
            if not recipe:
                continue

            ingredients = recipe.get('ingredients', [])
            for ingr in ingredients:
                try:
                    provider = Provider.objects.get(id=int(ingr.get('id')))
                    RecipeIngredient.objects.update_or_create(
                        enginery=enginery,
                        provider=provider,
                        ingredient_type=ingr.get('type', ''),
                        index=ingr.get('index'),
                        recipe_type=recipe.get('type', '')
                    )
                except Provider.DoesNotExist:
                    print(f"\n Can`n set RecipeIngredient for {enginery} as Provider({ingr.get('id')})")
                except Exception as exc:
                    print(f'\n {enginery}; Ingredient: {ingr}; {provider}', str(exc))
