import json
import os
import sys

from django.core.management import call_command
from django.test import TestCase

from spread_dashboard.project import models
from spread_dashboard.project import configs
from spread_dashboard.scheduler import models as sch_models


DATE = "1970-01-01"
HOLIDAYS = [DATE]
TASKS = [
    {
        "id": 17099,
        "enabled": False,
        "name": "\u0421\u0435\u0440\u0432\u0435\u0440\u043d\u0430\u044f",
        "triggers": [
            {
                "id": 21,
                "type": "workday",
                "time": "12:38:00",
                "actions": [
                    {
                        "id": 23,
                        "type": "broker",
                        "topic": "Spread/Set/1315/Location/333762/Lighting/SceneOn",
                        "value": 1
                    }
                ]
            }
        ]
    }
]

LIGHT_FACTORS = [
    {
        "value": 1.11,
        "enginery_id": 1337,
        "area_id": 1338
    }
]

STRATEGIES = [
    {
        "area_id": 1337,
        "natural_light": True,
        "load_balancing": True,
    },
]

CTP_FILE = 'throne-project_1104.ctp'


class FillCommandTestCase(TestCase):



    @classmethod
    def setUpClass(cls):
        sys.argv.append('fill')  # хак, чтоб адекватно отработали методы моделей
        configs.Projector.ctp_file = os.path.join(os.getcwd(), 'spread_core', 'tests', 'projector', CTP_FILE)
        _setup_files()

    @classmethod
    def tearDownClass(cls):
        sys.argv.remove('fill')
        _clean_files()

    def test_read_ctp(self):
        call_command('fill', '--ctp')
        item = models.Location.objects.filter(id=17099)
        self.assertTrue(bool(item), 'Не заполняется Location')
        item = models.Subginery.objects.get(id=171132)
        self.assertTrue(bool(item), 'Не заполняется Subginery')
        item = models.Manager.objects.get(id=4770)
        self.assertTrue(bool(item), 'Не заполняется Manager')
        item = models.Enginery.objects.get(id=171102)
        self.assertTrue(bool(item), 'Не заполняется Enginery')
        item = models.Provider.objects.get(id=178975)
        self.assertTrue(bool(item), 'Не заполняется Provider')
        item = models.RecipeIngredient.objects.filter(provider_id=178975)
        self.assertEquals(len(item), 2, 'Не заполняется RecipeIngredient')

    def test_fill_tasks(self):
        models.Location.objects.update_or_create(id=17099, name='test', scheduled=1, key="1", attributes="")
        call_command('fill', '--tasks')
        items = sch_models.SchLocation.objects.filter(id=17099)
        self.assertTrue(bool(items), "Не заполняется SchLocation")
        items = sch_models.SchTrigger.objects.filter(id=21)
        self.assertTrue(bool(items), "Не заполняется SchTrigger")
        items = sch_models.SchAction.objects.filter(id=23)
        self.assertTrue(bool(items), "Не заполняется SchAction")

    def test_fill_calendar(self):
        call_command('fill', '--calendar')
        items = sch_models.Holiday.objects.filter(date=DATE)
        self.assertTrue(bool(items))

    def test_fill_light_factors(self):
        models.Enginery.objects.create(id=1337)  # enginery_id
        models.Enginery.objects.create(id=1338)  # area_id
        call_command('fill', '--lightfactors')
        items = sch_models.LightFactor.objects.filter(enginery_id=1337, area_id=1338)
        self.assertTrue(bool(items))

    def test_fill_light_factors_enginery_not_exist(self):
        call_command('fill', '--lightfactors')
        items = sch_models.LightFactor.objects.filter(enginery_id=1337, area_id=1338)
        self.assertFalse(bool(items))

    def test_fill_strategies(self):
        models.Enginery.objects.create(id=1337)
        call_command('fill', '--strategies')
        items = sch_models.CommissioningStrategy.objects.filter(area_id=1337)
        self.assertTrue(bool(items))

    def test_fill_strategies_area_not_exists(self):
        call_command('fill', '--strategies')
        items = sch_models.CommissioningStrategy.objects.filter(area_id=1337)
        self.assertFalse(bool(items))


def _setup_files():
    dirname = os.path.dirname(configs.Dashboard().calendar)
    if not os.path.exists(dirname):
        os.makedirs(dirname, mode=0o700, exist_ok=True)
    with open(configs.Dashboard().calendar, 'w', encoding='utf-8') as file_:
        json.dump(HOLIDAYS, file_)
    with open(configs.Dashboard().tasks, 'w', encoding='utf-8') as file_:
        json.dump(TASKS, file_)
    with open(configs.Dashboard().light_factors, 'w', encoding='utf-8') as file_:
        json.dump(LIGHT_FACTORS, file_)
    with open(configs.Dashboard().area_strategies, 'w', encoding='utf-8') as file_:
        json.dump(STRATEGIES, file_)


def _clean_files():
    for file_ in (
            configs.Dashboard().calendar,
            configs.Dashboard().tasks,
            configs.Dashboard().light_factors,
            configs.Dashboard().area_strategies,
    ):
        os.remove(file_)
    dirname = os.path.dirname(configs.Dashboard().calendar)
    # удаляем все папки вверх по пути, остановимся нанепустой папке
    os.removedirs(dirname)
